"""
Key bindings.
"""
from __future__ import unicode_literals
from prompt_toolkit.enums import IncrementalSearchDirection
from prompt_toolkit.filters import HasFocus, Condition, HasSelection
from prompt_toolkit.key_binding.manager import KeyBindingManager as pt_KeyBindingManager
from prompt_toolkit.key_binding.vi_state import InputMode
from prompt_toolkit.keys import Keys
from prompt_toolkit.selection import SelectionType

from .enums import COMMAND, PROMPT
from .filters import WaitsForConfirmation, HasPrefix, InScrollBuffer, InScrollBufferNotSearching, InScrollBufferSearching
from .key_mappings import pymux_key_to_prompt_toolkit_key_sequence
from .commands.commands import call_command_handler

import six

__all__ = (
    'KeyBindingsManager',
)


class KeyBindingsManager(object):
    """
    Pymux key binding manager.
    """
    def __init__(self, pymux):
        self.pymux = pymux

        def get_search_state(cli):
            " Return the currently active SearchState. (The one for the focussed pane.) "
            return pymux.arrangement.get_active_pane(cli).search_state

        # Start from this KeyBindingManager from prompt_toolkit, to have basic
        # editing functionality for the command line. These key binding are
        # however only active when the following `enable_all` condition is met.
        self.pt_key_bindings_manager = pt_KeyBindingManager(
            enable_all=(HasFocus(COMMAND) | HasFocus(PROMPT) | InScrollBuffer(pymux)) & ~HasPrefix(pymux),
            enable_auto_suggest_bindings=True,
            enable_search=False,  # We have our own search bindings, that support multiple panes.
            enable_extra_page_navigation=True,
            get_search_state=get_search_state)

        self.registry = self.pt_key_bindings_manager.registry

        self._prefix = (Keys.ControlB, )
        self._prefix_binding = None

        # Load initial bindings.
        self._load_builtins()
        self._load_prefix_binding()
        _load_search_bindings(pymux, self.registry)

        # Custom user configured key bindings.
        # { (needs_prefix, key) -> (command, handler) }
        self.custom_bindings = {}

    def _load_prefix_binding(self):
        """
        Load the prefix key binding.
        """
        pymux = self.pymux
        registry = self.registry

        # Remove previous binding.
        if self._prefix_binding:
            self.registry.remove_binding(self._prefix_binding)

        # Create new Python binding.
        @registry.add_binding(*self._prefix, filter=
            ~(HasPrefix(pymux) | HasFocus(COMMAND) | HasFocus(PROMPT) | WaitsForConfirmation(pymux)))
        def enter_prefix_handler(event):
            " Enter prefix mode. "
            pymux.get_client_state(event.cli).has_prefix = True

        self._prefix_binding = enter_prefix_handler

    @property
    def prefix(self):
        " Get the prefix key. "
        return self._prefix

    @prefix.setter
    def prefix(self, keys):
        """
        Set a new prefix key.
        """
        assert isinstance(keys, tuple)

        self._prefix = keys
        self._load_prefix_binding()

    def _load_builtins(self):
        """
        Fill the Registry with the hard coded key bindings.
        """
        pymux = self.pymux
        registry = self.registry

        # Create filters.
        has_prefix = HasPrefix(pymux)
        waits_for_confirmation = WaitsForConfirmation(pymux)
        prompt_or_command_focus = HasFocus(COMMAND) | HasFocus(PROMPT)
        display_pane_numbers = Condition(lambda cli: pymux.display_pane_numbers)
        in_scroll_buffer_not_searching = InScrollBufferNotSearching(pymux)
        pane_input_allowed = ~(prompt_or_command_focus | has_prefix |
                               waits_for_confirmation | display_pane_numbers |
                               InScrollBuffer(pymux))

        @registry.add_binding(Keys.Any, filter=pane_input_allowed, invalidate_ui=False)
        def _(event):
            """
            When a pane has the focus, key bindings are redirected to the
            process running inside the pane.
            """
            # NOTE: we don't invalidate the UI, because for pymux itself,
            #       nothing in the output changes yet. It's the application in
            #       the pane that will probably echo back the typed characters.
            #       When we receive them, they are draw to the UI and it's
            #       invalidated.
            w = pymux.arrangement.get_active_window(event.cli)
            pane = w.active_pane

            if pane.clock_mode:
                # Leave clock mode on key press.
                pane.clock_mode = False
                pymux.invalidate()
            else:
                # Write input to pane. If 'synchronize_panes' is on, write
                # input to all panes in the current window.
                panes = w.panes if w.synchronize_panes else [pane]
                for p in panes:
                    p.process.write_key(event.key_sequence[0].key)

        @registry.add_binding(Keys.BracketedPaste, filter=pane_input_allowed, invalidate_ui=False)
        def _(event):
            """
            Pasting to the active pane. (Using bracketed paste.)
            """
            w = pymux.arrangement.get_active_window(event.cli)
            pane = w.active_pane

            if not pane.clock_mode:
                # Paste input to pane. If 'synchronize_panes' is on, paste
                # input to all panes in the current window.
                panes = w.panes if w.synchronize_panes else [pane]
                for p in panes:
                    p.process.write_input(event.data, paste=True)

        @registry.add_binding(Keys.Any, filter=has_prefix)
        def _(event):
            " Ignore unknown Ctrl-B prefixed key sequences. "
            pymux.get_client_state(event.cli).has_prefix = False

        @registry.add_binding(Keys.ControlC, filter=prompt_or_command_focus & ~has_prefix)
        @registry.add_binding(Keys.ControlG, filter=prompt_or_command_focus & ~has_prefix)
        @registry.add_binding(Keys.Backspace, filter=HasFocus(COMMAND) & ~has_prefix &
                              Condition(lambda cli: cli.buffers[COMMAND].text == ''))
        def _(event):
            " Leave command mode. "
            pymux.leave_command_mode(event.cli, append_to_history=False)

        @registry.add_binding('y', filter=waits_for_confirmation)
        @registry.add_binding('Y', filter=waits_for_confirmation)
        def _(event):
            """
            Confirm command.
            """
            client_state = pymux.get_client_state(event.cli)

            command = client_state.confirm_command
            client_state.confirm_command = None
            client_state.confirm_text = None

            pymux.handle_command(event.cli, command)

        @registry.add_binding('n', filter=waits_for_confirmation)
        @registry.add_binding('N', filter=waits_for_confirmation)
        @registry.add_binding(Keys.ControlC, filter=waits_for_confirmation)
        def _(event):
            """
            Cancel command.
            """
            client_state = pymux.get_client_state(event.cli)
            client_state.confirm_command = None
            client_state.confirm_text = None

        @registry.add_binding(Keys.ControlC, filter=in_scroll_buffer_not_searching)
        @registry.add_binding(Keys.ControlJ, filter=in_scroll_buffer_not_searching)
        @registry.add_binding('q', filter=in_scroll_buffer_not_searching)
        def _(event):
            " Exit scroll buffer. "
            pane = pymux.arrangement.get_active_pane(event.cli)
            pane.exit_scroll_buffer()

        @registry.add_binding(' ', filter=in_scroll_buffer_not_searching)
        def _(event):
            " Enter selection mode when pressing space in copy mode. "
            event.current_buffer.start_selection(selection_type=SelectionType.CHARACTERS)

        @registry.add_binding(Keys.ControlJ, filter=in_scroll_buffer_not_searching & HasSelection())
        def _(event):
            " Copy selection when pressing Enter. "
            clipboard_data = event.current_buffer.copy_selection()
            event.cli.clipboard.set_data(clipboard_data)

        @registry.add_binding('v', filter=in_scroll_buffer_not_searching & HasSelection())
        def _(event):
            " Toggle between selection types. "
            types = [SelectionType.LINES, SelectionType.BLOCK, SelectionType.CHARACTERS]
            selection_state = event.current_buffer.selection_state

            try:
                index = types.index(selection_state.type)
            except ValueError:  # Not in list.
                index = 0

            selection_state.type = types[(index + 1) % len(types)]

        @registry.add_binding(Keys.Any, filter=display_pane_numbers)
        def _(event):
            " When the pane numbers are shown. Any key press should hide them. "
            pymux.display_pane_numbers = False

        return registry

    def add_custom_binding(self, key_name, command, arguments, needs_prefix=False):
        """
        Add custom binding (for the "bind-key" command.)
        Raises ValueError if the give `key_name` is an invalid name.

        :param key_name: Pymux key name, for instance "C-a" or "M-x".
        """
        assert isinstance(key_name, six.text_type)
        assert isinstance(command, six.text_type)
        assert isinstance(arguments, list)

        # Unbind previous key.
        self.remove_custom_binding(key_name, needs_prefix=needs_prefix)

        # Translate the pymux key name into a prompt_toolkit key sequence.
        # (Can raise ValueError.)
        keys_sequence = pymux_key_to_prompt_toolkit_key_sequence(key_name)

        # Create handler and add to Registry.
        if needs_prefix:
            filter = HasPrefix(self.pymux)
        else:
            filter = ~HasPrefix(self.pymux)

        filter = filter & ~(WaitsForConfirmation(self.pymux) |
                            HasFocus(COMMAND) | HasFocus(PROMPT))

        def key_handler(event):
            " The actual key handler. "
            call_command_handler(command, self.pymux, event.cli, arguments)
            self.pymux.get_client_state(event.cli).has_prefix = False

        self.registry.add_binding(*keys_sequence, filter=filter)(key_handler)

        # Store key in `custom_bindings` in order to be able to call
        # "unbind-key" later on.
        k = (needs_prefix, key_name)
        self.custom_bindings[k] = CustomBinding(key_handler, command, arguments)

    def remove_custom_binding(self, key_name, needs_prefix=False):
        """
        Remove custom key binding for a key.

        :param key_name: Pymux key name, for instance "C-A".
        """
        k = (needs_prefix, key_name)

        if k in self.custom_bindings:
            self.registry.remove_binding(self.custom_bindings[k].handler)
            del self.custom_bindings[k]


class CustomBinding(object):
    """
    Record for storing a single custom key binding.
    """
    def __init__(self, handler, command, arguments):
        assert callable(handler)
        assert isinstance(command, six.text_type)
        assert isinstance(arguments, list)

        self.handler = handler
        self.command = command
        self.arguments = arguments


def _load_search_bindings(pymux, registry):
    """
    Load the key bindings for searching. (Vi and Emacs)

    This is different from the ones of prompt_toolkit, because we have a
    individual search buffers for each pane.
    """
    is_searching = InScrollBufferSearching(pymux)
    in_scroll_buffer_not_searching = InScrollBufferNotSearching(pymux)

    def search_buffer_is_empty(cli):
        """ Returns True when the search buffer is empty. """
        return pymux.arrangement.get_active_pane(cli).search_buffer.text == ''

    @registry.add_binding(Keys.ControlG, filter=is_searching)
    @registry.add_binding(Keys.ControlC, filter=is_searching)
    @registry.add_binding(Keys.Backspace, filter=is_searching & Condition(search_buffer_is_empty))
    def _(event):
        """
        Abort an incremental search and restore the original line.
        """
        pane = pymux.arrangement.get_active_pane(event.cli)
        pane.search_buffer.reset()
        pane.is_searching = False

    @registry.add_binding(Keys.ControlJ, filter=is_searching)
    def _(event):
        """
        When enter pressed in isearch, accept search.
        """
        pane = pymux.arrangement.get_active_pane(event.cli)

        input_buffer = pane.scroll_buffer
        search_buffer = pane.search_buffer

        # Update search state.
        if search_buffer.text:
            pane.search_state.text = search_buffer.text

        # Apply search.
        input_buffer.apply_search(pane.search_state, include_current_position=True)

        # Add query to history of search line.
        search_buffer.append_to_history()

        # Focus previous document again.
        pane.search_buffer.reset()
        pane.is_searching = False

    def enter_search(cli):
        cli.vi_state.input_mode = InputMode.INSERT

        pane = pymux.arrangement.get_active_pane(cli)
        pane.is_searching = True
        return pane.search_state

    @registry.add_binding(Keys.ControlR, filter=in_scroll_buffer_not_searching)
    @registry.add_binding('?', filter=in_scroll_buffer_not_searching)
    def _(event):
        " Enter reverse search. "
        search_state = enter_search(event.cli)
        search_state.direction = IncrementalSearchDirection.BACKWARD

    @registry.add_binding(Keys.ControlS, filter=in_scroll_buffer_not_searching)
    @registry.add_binding('/', filter=in_scroll_buffer_not_searching)
    def _(event):
        " Enter forward search. "
        search_state = enter_search(event.cli)
        search_state.direction = IncrementalSearchDirection.FORWARD

    @registry.add_binding(Keys.ControlR, filter=is_searching)
    @registry.add_binding(Keys.Up, filter=is_searching)
    def _(event):
        " Repeat reverse search. (While searching.) "
        pane = pymux.arrangement.get_active_pane(event.cli)

        # Update search_state.
        search_state = pane.search_state
        direction_changed = search_state.direction != IncrementalSearchDirection.BACKWARD

        search_state.text = pane.search_buffer.text
        search_state.direction = IncrementalSearchDirection.BACKWARD

        # Apply search to current buffer.
        if not direction_changed:
            pane.scroll_buffer.apply_search(
                pane.search_state, include_current_position=False, count=event.arg)

    @registry.add_binding(Keys.ControlS, filter=is_searching)
    @registry.add_binding(Keys.Down, filter=is_searching)
    def _(event):
        " Repeat forward search. (While searching.) "
        pane = pymux.arrangement.get_active_pane(event.cli)

        # Update search_state.
        search_state = pane.search_state
        direction_changed = search_state.direction != IncrementalSearchDirection.FORWARD

        search_state.text = pane.search_buffer.text
        search_state.direction = IncrementalSearchDirection.FORWARD

        # Apply search to current buffer.
        if not direction_changed:
            pane.scroll_buffer.apply_search(
                pane.search_state, include_current_position=False, count=event.arg)



"""
The color scheme.
"""
from __future__ import unicode_literals
from prompt_toolkit.styles import style_from_dict, Style, Attrs
from prompt_toolkit.styles.utils import split_token_in_parts, merge_attrs
from prompt_toolkit.token import Token

__all__ = (
    'PymuxStyle',
)


ui_style = {
    Token.Line:                         '#888888',
    Token.Line.Focussed:                '#448844',

    Token.TitleBar:                     'bg:#888888 #dddddd ',
    Token.TitleBar.Title:               '',
    Token.TitleBar.Name:                '#ffffff noitalic',
    Token.TitleBar.Name.Focussed:       'bg:#88aa44',
    Token.TitleBar.Line:                '#444444',
    Token.TitleBar.Line.Focussed:       '#448844 noinherit',
    Token.TitleBar.Focussed:            'bg:#5f875f #ffffff bold',
    Token.TitleBar.Focussed.Title:      '',
    Token.TitleBar.Zoom:                'bg:#884400 #ffffff',
    Token.TitleBar.PaneIndex:           '',
    Token.TitleBar.CopyMode:            'bg:#88aa88 #444444',
    Token.TitleBar.CopyMode.Position:   '',

    Token.TitleBar.Focussed.PaneIndex:         'bg:#88aa44 #ffffff',
    Token.TitleBar.Focussed.CopyMode:          'bg:#aaff44 #000000',
    Token.TitleBar.Focussed.CopyMode.Position: '#888888',

    Token.CommandLine:                  'bg:#4e4e4e #ffffff',
    Token.CommandLine.Command:          'bold',
    Token.CommandLine.Prompt:           'bold',
    Token.StatusBar:                    'bg:#444444 #ffffff',
    Token.StatusBar.Window:             'bg:#888888',
    Token.StatusBar.Window.Current:     '#88ff88 bold',
    Token.AutoSuggestion:               'bg:#4e5e4e #88aa88',
    Token.Message:                      'bg:#bbee88 #222222',
    Token.Background:                   '#888888',
    Token.Clock:                        'bg:#88aa00',
    Token.PaneNumber:                   'bg:#888888',
    Token.PaneNumber.Focussed:          'bg:#aa8800',
    Token.Terminated:                   'bg:#aa0000 #ffffff',

    Token.ConfirmationToolbar:          'bg:#880000 #ffffff',
    Token.ConfirmationToolbar.Question: '',
    Token.ConfirmationToolbar.YesNo:    'bg:#440000',

    Token.Search:                       'bg:#88aa88 #444444',
    Token.Search.Text:                  '',
    Token.Search.Focussed:              'bg:#aaff44 #444444',
    Token.Search.Focussed.Text:         'bold #000000',

    Token.SearchMatch:                  '#000000 bg:#88aa88',
    Token.SearchMatch.Current:          '#000000 bg:#aaffaa underline',


    # Completions menu.
    Token.Menu.Completions.Completion:         'bg:#88aa88 #222222',
    Token.Menu.Completions.Completion.Current: 'bg:#88cc88 #000000',
    Token.Menu.Completions.ProgressBar:        'bg:#889988',
    Token.Menu.Completions.ProgressButton:     'bg:#004400',
}


class PymuxStyle(Style):
    """
    The styling. It includes the UI style from above. But further, in order to
    proxy all the output from the processes, it interprets all tokens starting
    with ('C,) as tokens that describe their own style.
    """
    def __init__(self):
        self.ui_style = style_from_dict(ui_style)
        self._token_to_attrs_dict = None

    def get_attrs_for_token(self, token):
        result = []
        for part in split_token_in_parts(token):
            result.append(self._get_attrs_for_token(part))
        return merge_attrs(result)

    def _get_attrs_for_token(self, token):
        if token and token[0] == 'C':
            # Token starts with ('C',). Token describes its own style.
            c, fg, bg, bold, underline, italic, blink, reverse = token
            return Attrs(fg, bg, bold, underline, italic, blink, reverse)
        else:
            # Take styles from UI style.
            return self.ui_style.get_attrs_for_token(token)

    def invalidation_hash(self):
        return None



"""
Tools for Darwin. (Mac OS X.)
"""
from __future__ import unicode_literals
from ctypes import cdll, pointer, c_uint, c_ubyte, c_ulong
from struct import pack

import six

__all__ = (
    'get_proc_info',
    'get_proc_name'
)

# Current Values as of El Capitan

# /usr/include/sys/sysctl.h
CTL_KERN = 1
KERN_PROC = 14
KERN_PROC_PID = 1
KERN_PROC_PGRP = 2

# /usr/include/sys/param.h
MAXCOMLEN = 16

# kinfo_proc (/usr/include/sys/sysctl.h)
#  \-> extern_proc (/usr/include/sys/proc.h)
P_COMM_OFFSET = 243

# Finding the type for PIDs was *interesting*
# pid_t in /usr/include/sys/types.h -> \
#   pid_t in /usr/include/sys/_types/_pid_t.h -> \
#   __darwin_pid_t -> /usr/include/sys/_types.h -> \
#   __int32_t

LIBC = None


def _init():
    """
    Initialize ctypes DLL link.
    """
    global LIBC

    if LIBC is None:
        try:
            LIBC = cdll.LoadLibrary('libc.dylib')
        except OSError:
            # On OS X El Capitan, the above doesn't work for some reason and we
            # have to explicitely mention the path.
            # See: https://github.com/ffi/ffi/issues/461
            LIBC = cdll.LoadLibrary('/usr/lib/libc.dylib')


def get_proc_info(pid):
    """
    Use sysctl to retrieve process info.
    """
    # Ensure that we have the DLL loaded.
    _init()

    # Request the length of the process data.
    mib = (c_uint * 4)(CTL_KERN, KERN_PROC, KERN_PROC_PID, pid)
    oldlen = c_ulong()
    oldlenp = pointer(oldlen)
    r = LIBC.sysctl(mib, len(mib), None, oldlenp, None, 0)
    if r:
        return

    # Request the process data.
    reslen = oldlen.value
    old = (c_ubyte * reslen)()
    oldp = pointer(old)
    r = LIBC.sysctl(mib, len(mib), old, oldlenp, None, 0)
    if r:
        return
    #assert oldlen.value <= reslen

    return old[:reslen]


def get_proc_name(pid):
    """
    Use sysctl to retrive process name.
    """
    proc_kinfo = get_proc_info(pid)
    if not proc_kinfo:
        return

    p_comm_range = proc_kinfo[P_COMM_OFFSET:P_COMM_OFFSET + MAXCOMLEN + 1]
    p_comm_raw = ''.join(six.unichr(c) for c in p_comm_range)
    p_comm = p_comm_raw.split('\0', 1)[0]

    return p_comm



# encoding: utf-8
"""
The layout engine. This builds the prompt_toolkit layout.
"""
from __future__ import unicode_literals

from prompt_toolkit.enums import IncrementalSearchDirection
from prompt_toolkit.filters import Condition, to_cli_filter
from prompt_toolkit.layout.containers import VSplit, HSplit, Window, FloatContainer, Float, ConditionalContainer, Container
from prompt_toolkit.layout.controls import TokenListControl, FillControl, BufferControl
from prompt_toolkit.layout.dimension import LayoutDimension
from prompt_toolkit.layout.dimension import LayoutDimension as D
from prompt_toolkit.layout.lexers import Lexer
from prompt_toolkit.layout.lexers import SimpleLexer
from prompt_toolkit.layout.menus import CompletionsMenu
from prompt_toolkit.layout.processors import BeforeInput, AfterInput, AppendAutoSuggestion, Processor, Transformation, HighlightSearchProcessor, HighlightSelectionProcessor
from prompt_toolkit.layout.prompt import DefaultPrompt
from prompt_toolkit.layout.screen import Char
from prompt_toolkit.layout.screen import Point
from prompt_toolkit.layout.toolbars import TokenListToolbar
from prompt_toolkit.mouse_events import MouseEvent
from prompt_toolkit.mouse_events import MouseEventTypes
from prompt_toolkit.token import Token

from six.moves import range

import pymux.arrangement as arrangement
import datetime
import six
import weakref

from .enums import COMMAND, PROMPT
from .filters import WaitsForConfirmation, WaitsForPrompt, InCommandMode
from .format import format_pymux_string
from .log import logger
from .screen import DEFAULT_TOKEN

__all__ = (
    'LayoutManager',
)


class Justify:
    " Justify enum for the status bar. "
    LEFT = 'left'
    CENTER = 'center'
    RIGHT = 'right'

    _ALL = [LEFT, CENTER, RIGHT]


class Background(Container):
    """
    Generate the background of dots, which becomes visible when several clients
    are attached and not all of them have the same size.

    (This is implemented as a Container, rather than a UIControl wrapped in a
    Window, because it can be done very effecient this way.)
    """
    def reset(self):
        pass

    def preferred_width(self, cli, max_available_width):
        return D()

    def preferred_height(self, cli, width, max_available_height):
        return D()

    def write_to_screen(self, cli, screen, mouse_handlers, write_position):
        " Fill the whole area of write_position with dots. "
        default_char = Char(' ', Token.Background)
        dot = Char('.', Token.Background)

        ypos = write_position.ypos
        xpos = write_position.xpos

        for y in range(ypos, ypos + write_position.height):
            row = screen.data_buffer[y]

            for x in range(xpos, xpos + write_position.width):
                row[x] = dot if (x + y) % 3 == 0 else default_char

    def walk(self, cli):
        return []


# Numbers for the clock and pane numbering.
_numbers = list(zip(*[  # (Transpose x/y.)
    ['#####', '    #', '#####', '#####', '#   #', '#####', '#####', '#####', '#####', '#####'],
    ['#   #', '    #', '    #', '    #', '#   #', '#    ', '#    ', '    #', '#   #', '#   #'],
    ['#   #', '    #', '#####', '#####', '#####', '#####', '#####', '    #', '#####', '#####'],
    ['#   #', '    #', '#    ', '    #', '    #', '    #', '#   #', '    #', '#   #', '    #'],
    ['#####', '    #', '#####', '#####', '    #', '#####', '#####', '    #', '#####', '#####'],
]))


def _draw_number(screen, x_offset, y_offset, number, token=Token.Clock,
                 transparent=False):
    " Write number at position. "
    fg = Char(' ', token)
    bg = Char(' ', Token)

    for y, row in enumerate(_numbers[number]):
        screen_row = screen.data_buffer[y + y_offset]
        for x, n in enumerate(row):
            if n == '#':
                screen_row[x + x_offset] = fg
            elif not transparent:
                screen_row[x + x_offset] = bg


class BigClock(Container):
    """
    Display a big clock.
    """
    WIDTH = 28
    HEIGHT = 5

    def __init__(self, on_click):
        assert callable(on_click)
        self.on_click = on_click

    def reset(self):
        pass

    def write_to_screen(self, cli, screen, mouse_handlers, write_position):
        xpos = write_position.xpos
        ypos = write_position.ypos

        # Erase background.
        bg = Char(' ', Token)

        for y in range(ypos, self.HEIGHT + ypos):
            row = screen.data_buffer[y]
            for x in range(xpos, xpos + self.WIDTH):
                row[x] = bg

        # Display time.
        now = datetime.datetime.now()
        _draw_number(screen, xpos + 0, ypos, now.hour // 10)
        _draw_number(screen, xpos + 6, ypos, now.hour % 10)
        _draw_number(screen, xpos + 16, ypos, now.minute // 10)
        _draw_number(screen, xpos + 23, ypos, now.minute % 10)

        # Add a colon
        screen.data_buffer[ypos + 1][xpos + 13] = Char(' ', Token.Clock)
        screen.data_buffer[ypos + 3][xpos + 13] = Char(' ', Token.Clock)

        screen.width = self.WIDTH
        screen.height = self.HEIGHT

        mouse_handlers.set_mouse_handler_for_range(
            x_min=xpos,
            x_max=xpos + write_position.width,
            y_min=ypos,
            y_max=ypos + write_position.height,
            handler=self._mouse_handler)

    def _mouse_handler(self, cli, mouse_event):
        " Click callback. "
        if mouse_event.event_type == MouseEventTypes.MOUSE_UP:
            self.on_click(cli)
        else:
            return NotImplemented

    def preferred_width(self, cli, max_available_width):
        return D.exact(BigClock.WIDTH)

    def preferred_height(self, cli, width, max_available_height):
        return D.exact(BigClock.HEIGHT)

    def walk(self, cli):
        yield self


class PaneNumber(Container):
    """
    Number of panes, to be drawn in the middle of the pane.
    """
    WIDTH = 5
    HEIGHT = 5

    def __init__(self, pymux, arrangement_pane):
        self.pymux = pymux
        self.arrangement_pane = arrangement_pane

    def reset(self):
        pass

    def _get_index(self, cli):
        window = self.pymux.arrangement.get_active_window(cli)
        try:
            return window.get_pane_index(self.arrangement_pane)
        except ValueError:
            return 0

    def preferred_width(self, cli, max_available_width):
        # Enough to display all the digits.
        return LayoutDimension.exact(6 * len('%s' % self._get_index(cli)) - 1)

    def preferred_height(self, cli, width, max_available_height):
        return LayoutDimension.exact(self.HEIGHT)

    def write_to_screen(self, cli, screen, mouse_handlers, write_position):
        if self.pymux.arrangement.get_active_pane(cli) == self.arrangement_pane:
            token = Token.PaneNumber.Focussed
        else:
            token = Token.PaneNumber

        for i, d in enumerate('%s' % (self._get_index(cli))):
            _draw_number(screen, write_position.xpos + i * 6, write_position.ypos,
                         int(d), token=token, transparent=True)

    def walk(self, cli):
        yield self


class SearchWindow(Window):
    """
    Display the search input in copy mode.
    """
    def __init__(self, pymux, arrangement_pane):
        assert isinstance(arrangement_pane, arrangement.Pane)

        def focussed(cli):
            return pymux.arrangement.get_active_pane(cli) == arrangement_pane

        def get_before_input(cli):
            if not arrangement_pane.is_searching:
                text = ''
            elif arrangement_pane.search_state.direction == IncrementalSearchDirection.BACKWARD:
                text = 'Search up: '
            else:
                text = 'Search down: '

            if focussed(cli):
                return [(Token.Search.Focussed, text)]
            else:
                return [(Token.Search, text)]

        def get_after_input(cli):
            if focussed(cli):
                return [(Token.Search.Focussed, ' ')]
            else:
                return []

        class SearchLexer(Lexer):
            " Color for the search string. "
            def lex_document(self, cli, document):
                def get_line(lineno):
                    text = document.lines[lineno]
                    if focussed(cli):
                        return [(Token.Search.Focussed.Text, text)]
                    else:
                        return [(Token.Search.Text, text)]
                return get_line

        super(SearchWindow, self).__init__(
            content=BufferControl(
                buffer_name='search-%i' % arrangement_pane.pane_id,
                input_processors=[BeforeInput(get_before_input), AfterInput(get_after_input)],
                lexer=SearchLexer(),
                default_char=Char(token=Token)),
            dont_extend_height=True)


class MessageToolbar(TokenListToolbar):
    """
    Pop-up (at the bottom) for showing error/status messages.
    """
    def __init__(self, pymux):
        def get_message(cli):
            # If there is a message to be shown for this client, show that.
            client_state = pymux.get_client_state(cli)

            if client_state.message:
                return client_state.message
            else:
                return ''

        def get_tokens(cli):
            message = get_message(cli)
            if message:
                return [
                    (Token.Message, message),
                    (Token.SetCursorPosition, ''),
                    (Token.Message, ' '),
                ]
            else:
                return []

        f = Condition(lambda cli: get_message(cli) is not None)

        super(MessageToolbar, self).__init__(get_tokens, filter=f, has_focus=f)


class LayoutManager(object):
    """
    The main layout class, that contains the whole Pymux layout.
    """
    def __init__(self, pymux):
        self.pymux = pymux
        self.layout = self._create_layout()

        # Keep track of render information.
        self.pane_write_positions = {}
        self.body_write_position = None

    def _create_select_window_handler(self, window):
        " Return a mouse handler that selects the given window when clicking. "
        def handler(cli, mouse_event):
            if mouse_event.event_type == MouseEventTypes.MOUSE_DOWN:
                self.pymux.arrangement.set_active_window(cli, window)
                self.pymux.invalidate()
            else:
                return NotImplemented  # Event not handled here.
        return handler

    def _get_status_tokens(self, cli):
        " The tokens for the status bar. "
        result = []

        # Display panes.
        for i, w in enumerate(self.pymux.arrangement.windows):
            if i > 0:
                result.append((Token.StatusBar, ' '))

            if w == self.pymux.arrangement.get_active_window(cli):
                token = Token.StatusBar.Window.Current
                format_str = self.pymux.window_status_current_format

            else:
                token = Token.StatusBar.Window
                format_str = self.pymux.window_status_format

            result.append((
                token,
                format_pymux_string(self.pymux, cli, format_str, window=w),
                self._create_select_window_handler(w)))

        return result

    def _get_status_left_tokens(self, cli):
        return [
            (Token.StatusBar,
             format_pymux_string(self.pymux, cli, self.pymux.status_left)),
        ]

    def _get_status_right_tokens(self, cli):
        return [
            (Token.StatusBar,
             format_pymux_string(self.pymux, cli, self.pymux.status_right)),
        ]

    def _status_align_right(self, cli):
        return self.pymux.status_justify == Justify.RIGHT

    def _status_align_center(self, cli):
        return self.pymux.status_justify == Justify.CENTER

    def _before_prompt_command_tokens(self, cli):
        client_state = self.pymux.get_client_state(cli)
        return [(Token.CommandLine.Prompt, '%s ' % (client_state.prompt_text, ))]

    def _create_layout(self):
        """
        Generate the main prompt_toolkit layout.
        """
        waits_for_confirmation = WaitsForConfirmation(self.pymux)
        waits_for_prompt = WaitsForPrompt(self.pymux)
        in_command_mode = InCommandMode(self.pymux)

        return FloatContainer(
            content=HSplit([
                # The main window.
                HighlightBorders(self, self.pymux, FloatContainer(
                    Background(),
                    floats=[
                        Float(get_width=lambda cli: self.pymux.get_window_size(cli).columns,
                              get_height=lambda cli: self.pymux.get_window_size(cli).rows,
                              content=TraceBodyWritePosition(self.pymux, DynamicBody(self.pymux)))
                    ])),
                # Status bar.
                ConditionalContainer(
                    content=VSplit([
                        # Left.
                        Window(
                            height=D.exact(1),
                            get_width=(lambda cli: D(max=self.pymux.status_left_length)),
                            dont_extend_width=True,
                            content=TokenListControl(
                                self._get_status_left_tokens,
                                default_char=Char(' ', Token.StatusBar))),
                        # List of windows in the middle.
                        Window(
                            height=D.exact(1),
                            content=TokenListControl(
                                self._get_status_tokens,
                                align_right=Condition(self._status_align_right),
                                align_center=Condition(self._status_align_center),
                                default_char=Char(' ', Token.StatusBar))),
                        # Right.
                        Window(
                            height=D.exact(1),
                            get_width=(lambda cli: D(max=self.pymux.status_right_length)),
                            dont_extend_width=True,
                            content=TokenListControl(
                                self._get_status_right_tokens,
                                align_right=True,
                                default_char=Char(' ', Token.StatusBar)))
                    ]),
                    filter=Condition(lambda cli: self.pymux.enable_status),
                )
            ]),
            floats=[
                Float(bottom=1, left=0, content=MessageToolbar(self.pymux)),
                Float(left=0, right=0, bottom=0, content=HSplit([
                    # Wait for confirmation toolbar.
                    ConditionalContainer(
                        content=Window(
                            height=D.exact(1),
                            content=ConfirmationToolbar(self.pymux),
                        ),
                        filter=waits_for_confirmation,
                    ),
                    # ':' prompt toolbar.
                    ConditionalContainer(
                        content=Window(
                            height=D(min=1),  # Can be more if the command is multiline.
                            dont_extend_height=True,
                            content=BufferControl(
                                buffer_name=COMMAND,
                                default_char=Char(' ', Token.CommandLine),
                                lexer=SimpleLexer(Token.CommandLine),
                                preview_search=True,
                                input_processors=[
                                    AppendAutoSuggestion(),
                                    DefaultPrompt(lambda cli:[(Token.CommandLine.Prompt, ':')]),
                                    HighlightSelectionProcessor(),
                                ])
                        ),
                        filter=in_command_mode,
                    ),
                    # Other command-prompt commands toolbar.
                    ConditionalContainer(
                        content=Window(
                            height=D.exact(1),
                            content=BufferControl(
                                buffer_name=PROMPT,
                                default_char=Char(' ', Token.CommandLine),
                                lexer=SimpleLexer(Token.CommandLine),
                                input_processors=[
                                    BeforeInput(self._before_prompt_command_tokens),
                                    AppendAutoSuggestion(),
                                    HighlightSelectionProcessor(),
                                ])
                        ),
                        filter=waits_for_prompt,
                    ),
                ])),
                Float(xcursor=True, ycursor=True, content=CompletionsMenu(max_height=12)),
            ]
        )


class ConfirmationToolbar(TokenListControl):
    """
    Window that displays the yes/no confirmation dialog.
    """
    def __init__(self, pymux):
        token = Token.ConfirmationToolbar

        def get_tokens(cli):
            client_state = pymux.get_client_state(cli)
            return [
                (token.Question, ' '),
                (token.Question, format_pymux_string(
                    pymux, cli, client_state.confirm_text or '')),
                (token.Question, ' '),
                (token.YesNo, '  y/n'),
                (Token.SetCursorPosition, ''),
                (token.YesNo, '  '),
            ]

        super(ConfirmationToolbar, self).__init__(get_tokens, default_char=Char(' ', token))


class DynamicBody(Container):
    """
    The dynamic part, which is different for each CLI (for each client). It
    depends on which window/pane is active.

    This makes it possible to have just one main layout class, and
    automatically rebuild the parts that change if the windows/panes
    arrangement changes, without doing any synchronisation.
    """
    def __init__(self, pymux):
        self.pymux = pymux
        self._bodies_for_clis = weakref.WeakKeyDictionary()  # Maps CLI to (hash, Container)

    def _get_body(self, cli):
        " Return the Container object for the current CLI. "
        new_hash = self.pymux.arrangement.invalidation_hash(cli)

        # Return existing layout if nothing has changed to the arrangement.
        if cli in self._bodies_for_clis:
            existing_hash, container = self._bodies_for_clis[cli]
            if existing_hash == new_hash:
                return container

        # The layout changed. Build a new layout when the arrangement changed.
        new_layout = self._build_layout(cli)
        self._bodies_for_clis[cli] = (new_hash, new_layout)
        return new_layout

    def _build_layout(self, cli):
        " Rebuild a new Container object and return that. "
        logger.info('Rebuilding layout.')
        active_window = self.pymux.arrangement.get_active_window(cli)

        # When zoomed, only show the current pane, otherwise show all of them.
        if active_window.zoom:
            return _create_container_for_process(self.pymux, active_window.active_pane, zoom=True)
        else:
            return _create_split(self.pymux, self.pymux.arrangement.get_active_window(cli).root)

    def reset(self):
        for invalidation_hash, body in self._bodies_for_clis.values():
            body.reset()

    def preferred_width(self, cli, max_available_width):
        body = self._get_body(cli)
        return body.preferred_width(cli, max_available_width)

    def preferred_height(self, cli, width, max_available_height):
        body = self._get_body(cli)
        return body.preferred_height(cli, width, max_available_height)

    def write_to_screen(self, cli, screen, mouse_handlers, write_position):
        body = self._get_body(cli)
        body.write_to_screen(cli, screen, mouse_handlers, write_position)

    def walk(self, cli):
        # (Required for prompt_toolkit.layout.utils.find_window_for_buffer_name.)
        body = self._get_body(cli)
        return body.walk(cli)


def _create_split(pymux, split):
    """
    Create a prompt_toolkit `Container` instance for the given pymux split.
    """
    assert isinstance(split, (arrangement.HSplit, arrangement.VSplit))

    is_vsplit = isinstance(split, arrangement.VSplit)
    is_hsplit = not is_vsplit

    content = []

    def vertical_line():
        " Draw a vertical line between windows. (In case of a vsplit) "
        char = '│'
        content.append(HSplit([
                ConditionalContainer(
                    content=Window(
                        width=D.exact(1), height=D.exact(1),
                        content=FillControl(char, token=Token.TitleBar.Line)),
                    filter=Condition(lambda cli: pymux.enable_pane_status),
                ),
                Window(width=D.exact(1),
                       content=FillControl(char, token=Token.Line))
            ]))

    def horizontal_line():
        char = '─'
        content.append(
            ConditionalContainer(
                content=Window(height=D.exact(1),
                               content=FillControl(char, token=Token.Line)),
                filter=Condition(lambda cli: not pymux.enable_pane_status)),
        )

    for i, item in enumerate(split):
        if isinstance(item, (arrangement.VSplit, arrangement.HSplit)):
            content.append(_create_split(pymux, item))
        elif isinstance(item, arrangement.Pane):
            content.append(_create_container_for_process(pymux, item))
        else:
            raise TypeError('Got %r' % (item,))

        last_item = i == len(split) - 1
        if not last_item:
            if is_vsplit:
                vertical_line()
            elif is_hsplit:
                horizontal_line()

    def get_average_weight():
        """ Calculate average weight of the children. Return 1 if none of
        the children has a weight specified yet. """
        weights = 0
        count = 0

        for i in split:
            if i in split.weights:
                weights += split.weights[i]
                count += 1

        if weights:
            return max(1, weights // count)
        else:
            return 1

    def get_dimensions(cli):
        """
        Return a list of LayoutDimension instances for this split.
        These dimensions will take the weight from the
        arrangement.VSplit/HSplit instances.
        """
        average_weight = get_average_weight()

        # Make sure that weight is distributed

        result = []
        for i, item in enumerate(split):
            result.append(D(weight=split.weights.get(item) or average_weight))

            # Add dimension for the vertical border.
            last_item = i == len(split) - 1
            if is_vsplit and not last_item:
                result.append(D.exact(1))
            elif is_hsplit and not last_item:
                if pymux.enable_pane_status:
                    result.append(D.exact(0))
                else:
                    result.append(D.exact(1))

        return result

    def report_dimensions_callback(cli, dimensions):
        """
        When the layout is rendered, store the actial dimensions as
        weights in the arrangement.VSplit/HSplit classes.

        This is required because when a pane is resized with an increase of +1,
        we want to be sure that this corresponds exactly with one row or
        column. So, that updating weights corresponds exactly 1/1 to updating
        the size of the panes.
        """
        sizes = []
        for i, size in enumerate(dimensions):
            if i % 2 == 0:
                sizes.append(size)

        for c, size in zip(split, sizes):
            split.weights[c] = size

    # Create prompt_toolkit Container.
    return_cls = VSplit if is_vsplit else HSplit

    return return_cls(content, get_dimensions=get_dimensions,
                      report_dimensions_callback=report_dimensions_callback)


class _UseCopyTokenListProcessor(Processor):
    """
    In order to allow highlighting of the copy region, we use a preprocessed
    list of (Token, text) tuples. This processor returns just that list for the
    given pane.
    """
    def __init__(self, arrangement_pane):
        self.arrangement_pane = arrangement_pane

    def apply_transformation(self, cli, document, lineno, source_to_display, tokens):
        tokens = self.arrangement_pane.copy_get_tokens_for_line(lineno)
        return Transformation(tokens[:])

    def invalidation_hash(self, cli, document):
        return document.text


def _create_container_for_process(pymux, arrangement_pane, zoom=False):
    """
    Create a `Container` with a titlebar for a process.
    """
    assert isinstance(arrangement_pane, arrangement.Pane)
    process = arrangement_pane.process

    def has_focus(cli):
        return pymux.arrangement.get_active_pane(cli) == arrangement_pane

    def get_titlebar_token(cli):
        return Token.TitleBar.Focussed if has_focus(cli) else Token.TitleBar

    def get_titlebar_name_token(cli):
        return Token.TitleBar.Name.Focussed if has_focus(cli) else Token.TitleBar.Name

    def get_title_tokens(cli):
        token = get_titlebar_token(cli)
        name_token = get_titlebar_name_token(cli)
        result = []

        if zoom:
            result.append((Token.TitleBar.Zoom, ' Z '))

        if process.is_terminated:
            result.append((Token.Terminated, ' Terminated '))

        # Scroll buffer info.
        if arrangement_pane.display_scroll_buffer:
            result.append((token.CopyMode, ' %s ' % arrangement_pane.scroll_buffer_title))

            # Cursor position.
            document = arrangement_pane.scroll_buffer.document
            result.append((token.CopyMode.Position, ' %i,%i ' % (
                document.cursor_position_row, document.cursor_position_col)))

        if arrangement_pane.name:
            result.append((name_token, ' %s ' % arrangement_pane.name))
            result.append((token, ' '))

        return result + [
            (token.Title, format_pymux_string(pymux, cli, ' #T ', pane=arrangement_pane))  # XXX: Make configurable.
        ]

    def get_pane_index(cli):
        token = get_titlebar_token(cli)

        try:
            w = pymux.arrangement.get_active_window(cli)
            index = w.get_pane_index(arrangement_pane)
        except ValueError:
            index = '/'

        return [(token.PaneIndex, '%3s ' % index)]

    def on_click(cli):
        " Click handler for the clock. When clicked, select this pane. "
        arrangement_pane.clock_mode = False
        pymux.arrangement.get_active_window(cli).active_pane = arrangement_pane
        pymux.invalidate()

    def set_focus(cli):
        pymux.arrangement.get_active_window(cli).active_pane = arrangement_pane
        pymux.invalidate()

    clock_is_visible = Condition(lambda cli: arrangement_pane.clock_mode)
    pane_numbers_are_visible = Condition(lambda cli: pymux.display_pane_numbers)

    return TracePaneWritePosition(
        pymux, arrangement_pane,
        content=HSplit([
            # The title bar.
            ConditionalContainer(
                content=VSplit([
                        Window(
                            height=D.exact(1),
                            content=TokenListControl(
                                get_title_tokens,
                                get_default_char=lambda cli: Char(' ', get_titlebar_token(cli)))
                        ),
                        Window(
                            height=D.exact(1),
                            width=D.exact(4),
                            content=TokenListControl(get_pane_index)
                        )
                    ]),
                filter=Condition(lambda cli: pymux.enable_pane_status)),
            # The pane content.
            FloatContainer(
                content=HSplit([
                    # The 'screen' of the pseudo terminal.
                    ConditionalContainer(
                        content=Vt100Window(
                            process=process,
                            has_focus=Condition(lambda cli: (
                                cli.current_buffer_name != COMMAND and
                                pymux.arrangement.get_active_pane(cli) == arrangement_pane)),
                            set_focus=set_focus,
                        ),
                        filter=~clock_is_visible & Condition(lambda cli: not arrangement_pane.display_scroll_buffer)),

                    # The copy/paste buffer.
                    ConditionalContainer(
                        content=Window(BufferControl(
                            buffer_name='pane-%i' % arrangement_pane.pane_id,
                            focus_on_click=True,
                            default_char=Char(token=Token),
                            preview_search=True,
                            get_search_state=lambda cli: arrangement_pane.search_state,
                            search_buffer_name='search-%i' % arrangement_pane.pane_id,
                            input_processors=[
                                _UseCopyTokenListProcessor(arrangement_pane),
                                HighlightSearchProcessor(
                                    search_buffer_name='search-%i' % arrangement_pane.pane_id,
                                    get_search_state=lambda cli: arrangement_pane.search_state,
                                    preview_search=True,
                                ),
                                HighlightSelectionProcessor(),
                            ],
                        ),
                            wrap_lines=False,
                        ),
                        filter=~clock_is_visible & Condition(lambda cli: arrangement_pane.display_scroll_buffer)
                    ),
                    # Search toolbar. (Displayed when this pane has the focus, and searching.)
                    ConditionalContainer(
                        content=SearchWindow(pymux, arrangement_pane),
                        filter=~clock_is_visible & Condition(lambda cli: arrangement_pane.is_searching)
                    ),
                    # The clock.
                    ConditionalContainer(
                        # Add a dummy VSplit/HSplit around the BigClock in order to center it.
                        # (Using a FloatContainer to do the centering doesn't work well, because
                        # the boundaries are not clipped when the parent is smaller.)
                        content=VSplit([
                            Window(_FillControl(on_click)),
                            HSplit([
                                Window(_FillControl(on_click)),
                                BigClock(on_click),
                                Window(_FillControl(on_click)),
                            ]),
                            Window(_FillControl(on_click)),
                        ], get_dimensions=lambda cli: [None, D.exact(BigClock.WIDTH), None]),

                        filter=clock_is_visible,
                    ),
                ]),
                # Pane numbers. (Centered.)
                floats=[
                    Float(content=ConditionalContainer(
                        content=PaneNumber(pymux, arrangement_pane),
                        filter=pane_numbers_are_visible)),
                ]
            )
        ])
    )


class _FillControl(FillControl):
    """
    Extension to `FillControl` with click handlers.
    """
    def __init__(self, click_callback=None):
        self.click_callback = click_callback
        super(_FillControl, self).__init__()

    def mouse_handler(self, cli, mouse_event):
        " Call callback on click. "
        if mouse_event.event_type == MouseEventTypes.MOUSE_UP and self.click_callback:
            self.click_callback(cli)
        else:
            return NotImplemented


class _ContainerProxy(Container):
    def __init__(self, content):
        self.content = content

    def reset(self):
        self.content.reset()

    def preferred_width(self, cli, max_available_width):
        return self.content.preferred_width(cli, max_available_width)

    def preferred_height(self, cli, width, max_available_height):
        return self.content.preferred_height(cli, width, max_available_height)

    def write_to_screen(self, cli, screen, mouse_handlers, write_position):
        self.content.write_to_screen(cli, screen, mouse_handlers, write_position)

    def walk(self, cli):
        return self.content.walk(cli)


_focussed_border_titlebar = Char('┃', Token.TitleBar.Line.Focussed)
_focussed_border_vertical = Char('┃', Token.Line.Focussed)
_focussed_border_horizontal = Char('━', Token.Line.Focussed)
_focussed_border_left_top = Char('┏', Token.Line.Focussed)
_focussed_border_right_top = Char('┓', Token.Line.Focussed)
_focussed_border_left_bottom = Char('┗', Token.Line.Focussed)
_focussed_border_right_bottom = Char('┛', Token.Line.Focussed)

_border_vertical = Char('│', Token.Line)
_border_horizontal = Char('─', Token.Line)
_border_left_bottom = Char('└', Token.Line)
_border_right_bottom = Char('┘', Token.Line)
_border_left_top = Char('┌', Token.Line)
_border_right_top = Char('┐', Token.Line)


class HighlightBorders(_ContainerProxy):
    """
    Highlight the active borders. Happens post rendering.

    (We highlight the active pane when the rendering of everything else is
    done, otherwise, rendering of panes on the right will replace the result of
    this one.
    """
    def __init__(self, layout_manager, pymux, content):
        _ContainerProxy.__init__(self, content)
        self.pymux = pymux
        self.layout_manager = layout_manager

    def write_to_screen(self, cli, screen, mouse_handlers, write_position):
        # Clear previous list of pane coordinates.
        self.layout_manager.pane_write_positions = {}   # XXX: Should be for each CLI individually!!!!!
        self.layout_manager.body_write_position = None

        # Render everything.
        _ContainerProxy.write_to_screen(self, cli, screen, mouse_handlers, write_position)

        # When rendering is done. Draw borders and highlight the borders of the
        # active pane.
        self._draw_borders(screen, write_position)

        try:
            pane_wp = self.layout_manager.pane_write_positions[
                self.pymux.arrangement.get_active_pane(cli)]
        except KeyError:
            pass
        else:
            self._highlight_active_pane(screen, pane_wp, write_position)

    def _draw_borders(self, screen, write_position):
        """
        Draw borders around the whole window. (When there is space.)
        """
        data_buffer = screen.data_buffer

        if self.layout_manager.body_write_position:
            wp = self.layout_manager.body_write_position

            # Bottom line.
            if wp.ypos + wp.height < write_position.ypos + write_position.height:
                row = data_buffer[wp.ypos + wp.height]

                for x in range(wp.xpos, wp.xpos + wp.width):
                    row[x] = _border_horizontal

                # Left/right bottom.
                data_buffer[wp.ypos + wp.height][wp.xpos - 1] = _border_left_bottom
                data_buffer[wp.ypos + wp.height][wp.xpos + wp.width] = _border_right_bottom

            # Top line. (When we don't show the pane status bar.)
            if not self.pymux.enable_pane_status and wp.ypos >= 1:
                row = data_buffer[wp.ypos - 1]

                for x in range(wp.xpos, wp.xpos + wp.width):
                    row[x] = _border_horizontal

                # Left/right top.
                data_buffer[wp.ypos - 1][wp.xpos - 1] = _border_left_top
                data_buffer[wp.ypos - 1][wp.xpos + wp.width] = _border_right_top

            # Left and right line.
            for y in range(wp.ypos + 1, wp.ypos + wp.height):
                data_buffer[y][wp.xpos - 1] = _border_vertical
                data_buffer[y][wp.xpos + wp.width] = _border_vertical

            # Left/right top
            data_buffer[wp.ypos][wp.xpos - 1] = _border_left_top
            data_buffer[wp.ypos][wp.xpos + wp.width] = _border_right_top

    def _highlight_active_pane(self, screen, pane_wp, write_position):
        " Highlight the current, active pane. "
        data_buffer = screen.data_buffer
        xpos, ypos, width, height = pane_wp.xpos, pane_wp.ypos, pane_wp.width, pane_wp.height

        xleft = xpos - 1
        xright = xpos + width

        # Borders that are touching the pane status bar line.
        if self.pymux.enable_pane_status:
            row = data_buffer[ypos]
            row[xleft] = _focussed_border_left_top
            row[xright] = _focussed_border_right_top

        # Left and right border.
        if self.pymux.enable_pane_status:
            start_ypos = ypos + 1
        else:
            start_ypos = ypos

        for y in range(start_ypos, ypos + height):
            row = data_buffer[y]
            row[xleft] = row[xright] = _focussed_border_vertical

        # Draw the bottom line. (Only when there is space.)
        if ypos + height < write_position.ypos + write_position.height:
            row = data_buffer[ypos + height]

            for x in range(xpos, xpos + width):
                # Don't overwrite the titlebar of a pane below.
                if row[x].token == Token.Line:
                    row[x] = _focussed_border_horizontal

            # Bottom corners.
            row[xpos - 1] = _focussed_border_left_bottom
            row[xpos + width] = _focussed_border_right_bottom

        # Draw the top line.
        if not self.pymux.enable_pane_status and ypos >= 1:
            row = data_buffer[ypos - 1]

            for x in range(xpos, xpos + width):
                # Don't overwrite the titlebar of a pane below.
                if row[x].token == Token.Line:
                    row[x] = _focussed_border_horizontal

            # Bottom corners.
            row[xpos - 1] = _focussed_border_left_top
            row[xpos + width] = _focussed_border_right_top


class TracePaneWritePosition(_ContainerProxy):
    " Trace the write position of this pane. "
    def __init__(self, pymux, arrangement_pane, content):
        _ContainerProxy.__init__(self, content)

        self.pymux = pymux
        self.arrangement_pane = arrangement_pane

    def write_to_screen(self, cli, screen, mouse_handlers, write_position):
        _ContainerProxy.write_to_screen(self, cli, screen, mouse_handlers, write_position)

        self.pymux.layout_manager.pane_write_positions[self.arrangement_pane] = write_position


class TraceBodyWritePosition(_ContainerProxy):
    " Trace the write position of the whole body. "
    def __init__(self, pymux, content):
        _ContainerProxy.__init__(self, content)
        self.pymux = pymux

    def write_to_screen(self, cli, screen, mouse_handlers, write_position):
        _ContainerProxy.write_to_screen(self, cli, screen, mouse_handlers, write_position)
        self.pymux.layout_manager.body_write_position = write_position


def focus_left(pymux, cli):
    " Move focus to the left. "
    _move_focus(pymux, cli,
                lambda wp: wp.xpos - 2,  # 2 in order to skip over the border.
                lambda wp: wp.ypos)


def focus_right(pymux, cli):
    " Move focus to the right. "
    _move_focus(pymux, cli,
                lambda wp: wp.xpos + wp.width + 1,
                lambda wp: wp.ypos)


def focus_down(pymux, cli):
    " Move focus down. "
    _move_focus(pymux, cli,
                lambda wp: wp.xpos,
                lambda wp: wp.ypos + wp.height + 2)
        # 2 in order to skip over the border. Only required when the
        # pane-status is not shown, but a border instead.


def focus_up(pymux, cli):
    " Move focus up. "
    _move_focus(pymux, cli,
                lambda wp: wp.xpos,
                lambda wp: wp.ypos - 2)


def _move_focus(pymux, cli, get_x, get_y):
    " Move focus of the active window. "
    window = pymux.arrangement.get_active_window(cli)

    try:
        write_pos = pymux.layout_manager.pane_write_positions[window.active_pane]
    except KeyError:
        pass
    else:
        x = get_x(write_pos)
        y = get_y(write_pos)

        # Look for the pane at this position.
        for pane, wp in pymux.layout_manager.pane_write_positions.items():
            if (wp.xpos <= x < wp.xpos + wp.width and
                    wp.ypos <= y < wp.ypos + wp.height):
                window.active_pane = pane
                return


class Vt100Window(Container):
    """
    Container that holds the VT100 control.
    """
    def __init__(self, process, has_focus, set_focus):
        self.process = process
        self.has_focus = to_cli_filter(has_focus)
        self.set_focus = set_focus

    def reset(self):
        pass

    def preferred_width(self, cli, max_available_width):
        return LayoutDimension()

    def preferred_height(self, cli, width, max_available_height):
        return LayoutDimension()

    def write_to_screen(self, cli, screen, mouse_handlers, write_position):
        """
        Write window to screen. This renders the user control, the margins and
        copies everything over to the absolute position at the given screen.
        """
        # Set size of the screen.
        self.process.set_size(write_position.width, write_position.height)

        vertical_scroll = self.process.screen.line_offset

        # Render UserControl.
        temp_screen = self.process.screen.pt_screen

        # Write body to screen.
        self._copy_body(cli, temp_screen, screen, write_position, vertical_scroll,
                        write_position.width)

        # Set mouse handlers.
        def mouse_handler(cli, mouse_event):
            """ Wrapper around the mouse_handler of the `UIControl` that turns
            absolute coordinates into relative coordinates. """
            position = mouse_event.position

            # Call the mouse handler of the UIControl first.
            self._mouse_handler(
                cli, MouseEvent(
                    position=Point(x=position.x - write_position.xpos,
                                   y=position.y - write_position.ypos + vertical_scroll),
                    event_type=mouse_event.event_type))

        mouse_handlers.set_mouse_handler_for_range(
            x_min=write_position.xpos,
            x_max=write_position.xpos + write_position.width,
            y_min=write_position.ypos,
            y_max=write_position.ypos + write_position.height,
            handler=mouse_handler)

        # If reverse video is enabled for the whole screen.
        if self.process.screen.has_reverse_video:
            data_buffer = screen.data_buffer

            for y in range(write_position.ypos, write_position.ypos + write_position.height):
                row = data_buffer[y]

                for x in range(write_position.xpos, write_position.xpos + write_position.width):
                    char = row[x]
                    token = list(char.token or DEFAULT_TOKEN)

                    # The token looks like ('C', *attrs). Replace the value of the reverse flag.
                    if token and token[0] == 'C':
                        token[-1] = not token[-1]  # Invert reverse value.
                        row[x] = Char(char.char, tuple(token))

    def _copy_body(self, cli, temp_screen, new_screen, write_position,
                   vertical_scroll, width):
        """
        Copy characters from the temp screen that we got from the `UIControl`
        to the real screen.
        """
        xpos = write_position.xpos
        ypos = write_position.ypos
        height = write_position.height

        temp_buffer = temp_screen.data_buffer
        new_buffer = new_screen.data_buffer
        temp_screen_height = temp_screen.height

        vertical_scroll = self.process.screen.line_offset
        y = 0

        # Now copy the region we need to the real screen.
        for y in range(0, height):
            # We keep local row variables. (Don't look up the row in the dict
            # for each iteration of the nested loop.)
            new_row = new_buffer[y + ypos]

            if y >= temp_screen_height and y >= write_position.height:
                # Break out of for loop when we pass after the last row of the
                # temp screen. (We use the 'y' position for calculation of new
                # screen's height.)
                break
            else:
                temp_row = temp_buffer[y + vertical_scroll]

                # Copy row content, except for transparent tokens.
                # (This is useful in case of floats.)
                for x in range(0, width):
                    new_row[x + xpos] = temp_row[x]

        if self.has_focus(cli):
            new_screen.cursor_position = Point(
                y=temp_screen.cursor_position.y + ypos - vertical_scroll,
                x=temp_screen.cursor_position.x + xpos)

            new_screen.show_cursor = temp_screen.show_cursor

        # Update height of the output screen. (new_screen.write_data is not
        # called, so the screen is not aware of its height.)
        new_screen.height = max(new_screen.height, ypos + y + 1)

    def _mouse_handler(self, cli, mouse_event):
        """
        Handle mouse events in a pane. A click in a non-active pane will select
        it, one in an active pane, will send the mouse event to the application
        running inside it.
        """
        process = self.process
        x = mouse_event.position.x
        y = mouse_event.position.y

        # The containing Window translates coordinates to the absolute position
        # of the whole screen, but in this case, we need the relative
        # coordinates of the visible area.
        y -= self.process.screen.line_offset

        if not self.has_focus(cli):
            # Focus this process when the mouse has been clicked.
            if mouse_event.event_type == MouseEventTypes.MOUSE_UP:
                self.set_focus(cli)
        else:
            # Already focussed, send event to application when it requested
            # mouse support.
            if process.screen.sgr_mouse_support_enabled:
                # Xterm SGR mode.
                ev, m = {
                    MouseEventTypes.MOUSE_DOWN: ('0', 'M'),
                    MouseEventTypes.MOUSE_UP: ('0', 'm'),
                    MouseEventTypes.SCROLL_UP: ('64', 'M'),
                    MouseEventTypes.SCROLL_DOWN: ('65', 'M'),
                }.get(mouse_event.event_type)

                self.process.write_input(
                    '\x1b[<%s;%s;%s%s' % (ev, x + 1, y + 1, m))

            elif process.screen.urxvt_mouse_support_enabled:
                # Urxvt mode.
                ev = {
                    MouseEventTypes.MOUSE_DOWN: 32,
                    MouseEventTypes.MOUSE_UP: 35,
                    MouseEventTypes.SCROLL_UP: 96,
                    MouseEventTypes.SCROLL_DOWN: 97,
                }.get(mouse_event.event_type)

                self.process.write_input(
                    '\x1b[%s;%s;%sM' % (ev, x + 1, y + 1))

            elif process.screen.mouse_support_enabled:
                # Fall back to old mode.
                if x < 96 and y < 96:
                    ev = {
                            MouseEventTypes.MOUSE_DOWN: 32,
                            MouseEventTypes.MOUSE_UP: 35,
                            MouseEventTypes.SCROLL_UP: 96,
                            MouseEventTypes.SCROLL_DOWN: 97,
                    }.get(mouse_event.event_type)

                    self.process.write_input('\x1b[M%s%s%s' % (
                        six.unichr(ev),
                        six.unichr(x + 33),
                        six.unichr(y + 33)))

    def walk(self, cli):
        # Only yield self. A window doesn't have children.
        yield self



"""
Pymux string formatting.
"""
from __future__ import unicode_literals
import datetime
import socket
import six

__all__ = (
    'format_pymux_string',
)


def format_pymux_string(pymux, cli, string, window=None, pane=None):
    """
    Apply pymux sting formatting. (Similar to tmux.)
    E.g.  #P is replaced by the index of the active pane.

    We try to stay compatible with tmux, if possible.
    One thing that we won't support (for now) is colors, because our styling
    works different. (With a Style class.) On the other hand, in the future, we
    could allow things like `#[token=Token.Title.PID]`. This gives a clean
    separation of semantics and colors, making it easy to write different color
    schemes.
    """
    arrangement = pymux.arrangement

    if window is None:
        window = arrangement.get_active_window(cli)

    if pane is None:
        pane = window.active_pane

    def id_of_pane():
        return '%s' % (pane.pane_id, )

    def index_of_pane():
        try:
            return '%s' % (window.get_pane_index(pane), )
        except ValueError:
            return '/'

    def index_of_window():
        return '%s' % (window.index, )

    def name_of_window():
        return window.name or '(noname)'

    def window_flags():
        z = 'Z' if window.zoom else ''

        if window == arrangement.get_active_window(cli):
            return '*' + z
        elif window == arrangement.get_previous_active_window(cli):
            return '-' + z
        else:
            return z + ' '

    def name_of_session():
        return pymux.session_name

    def title_of_pane():
        return pane.process.screen.title

    def hostname():
        return socket.gethostname()

    def literal():
        return '#'

    format_table = {
        '#D': id_of_pane,
        '#F': window_flags,
        '#I': index_of_window,
        '#P': index_of_pane,
        '#S': name_of_session,
        '#T': title_of_pane,
        '#W': name_of_window,
        '#h': hostname,
        '##': literal,
    }

    # Date/time formatting.
    if '%' in string:
        try:
            if six.PY2:
                string = datetime.datetime.now().strftime(
                    string.encode('utf-8')).decode('utf-8')
            else:
                string = datetime.datetime.now().strftime(string)
        except ValueError:  # strftime format ends with raw %
            string = '<ValueError>'

    # Apply '#' formatting.
    for symbol, f in format_table.items():
        if symbol in string:
            string = string.replace(symbol, f())

    return string



"""
All configurable options which can be changed through "set-option" commands.
"""
from __future__ import unicode_literals
from abc import ABCMeta, abstractmethod
import six

from .key_mappings import PYMUX_TO_PROMPT_TOOLKIT_KEYS, pymux_key_to_prompt_toolkit_key_sequence
from .utils import get_default_shell
from .layout import Justify

__all__ = (
    'Option',
    'SetOptionError',
    'OnOffOption',
    'ALL_OPTIONS',
    'ALL_WINDOW_OPTIONS',
)


class Option(six.with_metaclass(ABCMeta, object)):
    """
    Base class for all options.
    """
    @abstractmethod
    def get_all_values(self):
        """
        Return a list of strings, with all possible values. (For
        autocompletion.)
        """

    @abstractmethod
    def set_value(self, pymux, cli, value):
        " Set option. This can raise SetOptionError. "


class SetOptionError(Exception):
    """
    Raised when setting an option fails.
    """
    def __init__(self, message):
        self.message = message


class OnOffOption(Option):
    """
    Boolean on/off option.
    """
    def __init__(self, attribute_name, window_option=False):
        self.attribute_name = attribute_name
        self.window_option = window_option

    def get_all_values(self, pymux):
        return ['on', 'off']

    def set_value(self, pymux, cli, value):
        value = value.lower()

        if value in ('on', 'off'):
            if self.window_option:
                w = pymux.arrangement.get_active_window(cli)
                setattr(w, self.attribute_name, (value == 'on'))
            else:
                setattr(pymux, self.attribute_name, (value == 'on'))
        else:
            raise SetOptionError('Expecting "yes" or "no".')


class StringOption(Option):
    """
    String option, the attribute is set as a Pymux attribute.
    """
    def __init__(self, attribute_name, possible_values=None):
        self.attribute_name = attribute_name
        self.possible_values = possible_values or []

    def get_all_values(self, pymux):
        return sorted(set(
            self.possible_values + [getattr(pymux, self.attribute_name)]
        ))

    def set_value(self, pymux, cli, value):
        setattr(pymux, self.attribute_name, value)


class PositiveIntOption(Option):
    """
    Positive integer option, the attribute is set as a Pymux attribute.
    """
    def __init__(self, attribute_name, possible_values=None):
        self.attribute_name = attribute_name
        self.possible_values = ['%s' % i for i in (possible_values or [])]

    def get_all_values(self, pymux):
        return sorted(set(
            self.possible_values +
            ['%s' % getattr(pymux, self.attribute_name)]
        ))

    def set_value(self, pymux, cli, value):
        """
        Take a string, and return an integer. Raise SetOptionError when the
        given text does not parse to a positive integer.
        """
        try:
            value = int(value)
            if value < 0:
                raise ValueError
        except ValueError:
            raise SetOptionError('Expecting an integer.')
        else:
            setattr(pymux, self.attribute_name, value)


class KeyPrefixOption(Option):
    def get_all_values(self, pymux):
        return PYMUX_TO_PROMPT_TOOLKIT_KEYS.keys()

    def set_value(self, pymux, cli, value):
        # Translate prefix to prompt_toolkit
        try:
            keys = pymux_key_to_prompt_toolkit_key_sequence(value)
        except ValueError:
            raise SetOptionError('Invalid key: %r' % (value, ))
        else:
            pymux.key_bindings_manager.prefix = keys


class BaseIndexOption(Option):
    " Base index for window numbering. "
    def get_all_values(self, pymux):
        return ['0', '1']

    def set_value(self, pymux, cli, value):
        try:
            value = int(value)
        except ValueError:
            raise SetOptionError('Expecting an integer.')
        else:
            pymux.arrangement.base_index = value


class KeysOption(Option):
    " Emacs or Vi mode. "
    def __init__(self, attribute_name):
        self.attribute_name = attribute_name

    def get_all_values(self, pymux):
        return ['emacs', 'vi']

    def set_value(self, pymux, cli, value):
        if value in ('emacs', 'vi'):
            setattr(pymux, self.attribute_name, value == 'vi')
        else:
            raise SetOptionError('Expecting "vi" or "emacs".')

class JustifyOption(Option):
    def __init__(self, attribute_name):
        self.attribute_name = attribute_name

    def get_all_values(self, pymux):
        return Justify._ALL

    def set_value(self, pymux, cli, value):
        if value in Justify._ALL:
            setattr(pymux, self.attribute_name, value)
        else:
            raise SetOptionError('Invalid justify option.')


ALL_OPTIONS = {
    'base-index': BaseIndexOption(),
    'bell': OnOffOption('enable_bell'),
    'history-limit': PositiveIntOption(
        'history_limit', [200, 500, 1000, 2000, 5000, 10000]),
    'mouse': OnOffOption('enable_mouse_support'),
    'prefix': KeyPrefixOption(),
    'remain-on-exit': OnOffOption('remain_on_exit'),
    'status': OnOffOption('enable_status'),
    'pane-border-status': OnOffOption('enable_pane_status'),
    'status-keys': KeysOption('status_keys_vi_mode'),
    'mode-keys': KeysOption('mode_keys_vi_mode'),
    'default-terminal': StringOption(
        'default_terminal', ['xterm', 'xterm-256color', 'screen']),
    'status-right': StringOption('status_right'),
    'status-left': StringOption('status_left'),
    'status-right-length': PositiveIntOption('status_right_length', [20]),
    'status-left-length': PositiveIntOption('status_left_length', [20]),
    'window-status-format': StringOption('window_status_format'),
    'window-status-current-format': StringOption('window_status_current_format'),
    'default-shell': StringOption(
        'default_shell', [get_default_shell()]),
    'status-justify': JustifyOption('status_justify'),
    'status-interval': PositiveIntOption(
        'status_interval', [1, 2, 4, 8, 16, 30, 60]),
}


ALL_WINDOW_OPTIONS = {
    'synchronize-panes': OnOffOption('synchronize_panes', window_option=True),
}



from __future__ import unicode_literals
from prompt_toolkit.filters import Filter

__all__ = (
    'HasPrefix',
    'WaitsForConfirmation',
    'InCommandMode',
    'WaitsForPrompt',
    'InScrollBuffer',
    'InScrollBufferNotSearching',
    'InScrollBufferSearching',
)


class HasPrefix(Filter):
    """
    When the prefix key (Usual C-b) has been pressed.
    """
    def __init__(self, pymux):
        self.pymux = pymux

    def __call__(self, cli):
        return self.pymux.get_client_state(cli).has_prefix


class WaitsForConfirmation(Filter):
    """
    Waiting for a yes/no key press.
    """
    def __init__(self, pymux):
        self.pymux = pymux

    def __call__(self, cli):
        return bool(self.pymux.get_client_state(cli).confirm_command)


class InCommandMode(Filter):
    """
    When ':' has been pressed.'
    """
    def __init__(self, pymux):
        self.pymux = pymux

    def __call__(self, cli):
        client_state = self.pymux.get_client_state(cli)
        return client_state.command_mode and not client_state.confirm_command


class WaitsForPrompt(Filter):
    """
    Waiting for input for a "command-prompt" command.
    """
    def __init__(self, pymux):
        self.pymux = pymux

    def __call__(self, cli):
        client_state = self.pymux.get_client_state(cli)
        return bool(client_state.prompt_command) and not client_state.confirm_command


def _confirm_or_prompt_or_command(pymux, cli):
    " True when we are waiting for a command, prompt or confirmation. "
    client_state = pymux.get_client_state(cli)
    if client_state.confirm_text or client_state.prompt_command or client_state.command_mode:
        return True


class InScrollBuffer(Filter):
    def __init__(self, pymux):
        self.pymux = pymux

    def __call__(self, cli):
        if _confirm_or_prompt_or_command(self.pymux, cli):
            return False

        pane = self.pymux.arrangement.get_active_pane(cli)
        return pane.display_scroll_buffer


class InScrollBufferNotSearching(Filter):
    def __init__(self, pymux):
        self.pymux = pymux

    def __call__(self, cli):
        if _confirm_or_prompt_or_command(self.pymux, cli):
            return False

        pane = self.pymux.arrangement.get_active_pane(cli)
        return pane.display_scroll_buffer and not pane.is_searching


class InScrollBufferSearching(Filter):
    def __init__(self, pymux):
        self.pymux = pymux

    def __call__(self, cli):
        if _confirm_or_prompt_or_command(self.pymux, cli):
            return False

        pane = self.pymux.arrangement.get_active_pane(cli)
        return pane.display_scroll_buffer and pane.is_searching



"""
Mapping between vt100 key sequences, the prompt_toolkit key constants and the
Pymux namings. (Those namings are kept compatible with tmux.)
"""
from __future__ import unicode_literals
from prompt_toolkit.keys import Keys
from prompt_toolkit.terminal.vt100_input import ANSI_SEQUENCES

__all__ = (
    'pymux_key_to_prompt_toolkit_key_sequence',
    'prompt_toolkit_key_to_vt100_key',
    'PYMUX_TO_PROMPT_TOOLKIT_KEYS',
)


def pymux_key_to_prompt_toolkit_key_sequence(key):
    """
    Turn a pymux description of a key. E.g.  "C-a" or "M-x" into a
    prompt-toolkit key sequence.

    Raises `ValueError` if the key is not known.
    """
    # Make the c- and m- prefixes case insensitive.
    if key.lower().startswith('m-c-'):
        key = 'M-C-' + key[4:]
    elif key.lower().startswith('c-'):
        key = 'C-' + key[2:]
    elif key.lower().startswith('m-'):
        key = 'M-' + key[2:]

    # Lookup key.
    try:
        return PYMUX_TO_PROMPT_TOOLKIT_KEYS[key]
    except KeyError:
        if len(key) == 1:
            return (key, )
        else:
            raise ValueError('Unknown key: %r' % (key, ))


# Create a mapping from prompt_toolkit keys to their ANSI sequences.
# TODO: This is not completely correct yet. It doesn't take
#       cursor/application mode into account. Create new tables for this.
_PROMPT_TOOLKIT_KEY_TO_VT100 = dict(
    (key, vt100_data) for vt100_data, key in ANSI_SEQUENCES.items())


def prompt_toolkit_key_to_vt100_key(key, application_mode=False):
    """
    Turn a prompt toolkit key. (E.g Keys.ControlB) into a Vt100 key sequence.
    (E.g. \x1b[A.)
    """
    application_mode_keys = {
        Keys.Up: '\x1bOA',
        Keys.Left: '\x1bOD',
        Keys.Right: '\x1bOC',
        Keys.Down: '\x1bOB',
    }

    if key == Keys.ControlJ:
        # Required for redis-cli. This can be removed when prompt_toolkit stops
        # replacing \r by \n.
        return '\r'

    if key == '\n':
        return '\r'

    elif application_mode and key in application_mode_keys:
        return application_mode_keys.get(key)
    else:
        return _PROMPT_TOOLKIT_KEY_TO_VT100.get(key, key)


PYMUX_TO_PROMPT_TOOLKIT_KEYS = {
    'Space': (' '),

    'C-a': (Keys.ControlA, ),
    'C-b': (Keys.ControlB, ),
    'C-c': (Keys.ControlC, ),
    'C-d': (Keys.ControlD, ),
    'C-e': (Keys.ControlE, ),
    'C-f': (Keys.ControlF, ),
    'C-g': (Keys.ControlG, ),
    'C-h': (Keys.ControlH, ),
    'C-i': (Keys.ControlI, ),
    'C-j': (Keys.ControlJ, ),
    'C-k': (Keys.ControlK, ),
    'C-l': (Keys.ControlL, ),
    'C-m': (Keys.ControlM, ),
    'C-n': (Keys.ControlN, ),
    'C-o': (Keys.ControlO, ),
    'C-p': (Keys.ControlP, ),
    'C-q': (Keys.ControlQ, ),
    'C-r': (Keys.ControlR, ),
    'C-s': (Keys.ControlS, ),
    'C-t': (Keys.ControlT, ),
    'C-u': (Keys.ControlU, ),
    'C-v': (Keys.ControlV, ),
    'C-w': (Keys.ControlW, ),
    'C-x': (Keys.ControlX, ),
    'C-y': (Keys.ControlY, ),
    'C-z': (Keys.ControlZ, ),

    'C-Left': (Keys.ControlLeft, ),
    'C-Right': (Keys.ControlRight, ),
    'C-Up': (Keys.ControlUp, ),
    'C-Down': (Keys.ControlDown, ),
    'C-\\': (Keys.ControlBackslash, ),

    'S-Left':  (Keys.ShiftLeft, ),
    'S-Right': (Keys.ShiftRight, ),
    'S-Up':    (Keys.ShiftUp, ),
    'S-Down':  (Keys.ShiftDown, ),

    'M-C-a': (Keys.Escape, Keys.ControlA, ),
    'M-C-b': (Keys.Escape, Keys.ControlB, ),
    'M-C-c': (Keys.Escape, Keys.ControlC, ),
    'M-C-d': (Keys.Escape, Keys.ControlD, ),
    'M-C-e': (Keys.Escape, Keys.ControlE, ),
    'M-C-f': (Keys.Escape, Keys.ControlF, ),
    'M-C-g': (Keys.Escape, Keys.ControlG, ),
    'M-C-h': (Keys.Escape, Keys.ControlH, ),
    'M-C-i': (Keys.Escape, Keys.ControlI, ),
    'M-C-j': (Keys.Escape, Keys.ControlJ, ),
    'M-C-k': (Keys.Escape, Keys.ControlK, ),
    'M-C-l': (Keys.Escape, Keys.ControlL, ),
    'M-C-m': (Keys.Escape, Keys.ControlM, ),
    'M-C-n': (Keys.Escape, Keys.ControlN, ),
    'M-C-o': (Keys.Escape, Keys.ControlO, ),
    'M-C-p': (Keys.Escape, Keys.ControlP, ),
    'M-C-q': (Keys.Escape, Keys.ControlQ, ),
    'M-C-r': (Keys.Escape, Keys.ControlR, ),
    'M-C-s': (Keys.Escape, Keys.ControlS, ),
    'M-C-t': (Keys.Escape, Keys.ControlT, ),
    'M-C-u': (Keys.Escape, Keys.ControlU, ),
    'M-C-v': (Keys.Escape, Keys.ControlV, ),
    'M-C-w': (Keys.Escape, Keys.ControlW, ),
    'M-C-x': (Keys.Escape, Keys.ControlX, ),
    'M-C-y': (Keys.Escape, Keys.ControlY, ),
    'M-C-z': (Keys.Escape, Keys.ControlZ, ),

    'M-C-Left': (Keys.Escape, Keys.ControlLeft, ),
    'M-C-Right': (Keys.Escape, Keys.ControlRight, ),
    'M-C-Up': (Keys.Escape, Keys.ControlUp, ),
    'M-C-Down': (Keys.Escape, Keys.ControlDown, ),
    'M-C-\\': (Keys.Escape, Keys.ControlBackslash, ),

    'M-a': (Keys.Escape, 'a'),
    'M-b': (Keys.Escape, 'b'),
    'M-c': (Keys.Escape, 'c'),
    'M-d': (Keys.Escape, 'd'),
    'M-e': (Keys.Escape, 'e'),
    'M-f': (Keys.Escape, 'f'),
    'M-g': (Keys.Escape, 'g'),
    'M-h': (Keys.Escape, 'h'),
    'M-i': (Keys.Escape, 'i'),
    'M-j': (Keys.Escape, 'j'),
    'M-k': (Keys.Escape, 'k'),
    'M-l': (Keys.Escape, 'l'),
    'M-m': (Keys.Escape, 'm'),
    'M-n': (Keys.Escape, 'n'),
    'M-o': (Keys.Escape, 'o'),
    'M-p': (Keys.Escape, 'p'),
    'M-q': (Keys.Escape, 'q'),
    'M-r': (Keys.Escape, 'r'),
    'M-s': (Keys.Escape, 's'),
    'M-t': (Keys.Escape, 't'),
    'M-u': (Keys.Escape, 'u'),
    'M-v': (Keys.Escape, 'v'),
    'M-w': (Keys.Escape, 'w'),
    'M-x': (Keys.Escape, 'x'),
    'M-y': (Keys.Escape, 'y'),
    'M-z': (Keys.Escape, 'z'),

    'M-0': (Keys.Escape, '0'),
    'M-1': (Keys.Escape, '1'),
    'M-2': (Keys.Escape, '2'),
    'M-3': (Keys.Escape, '3'),
    'M-4': (Keys.Escape, '4'),
    'M-5': (Keys.Escape, '5'),
    'M-6': (Keys.Escape, '6'),
    'M-7': (Keys.Escape, '7'),
    'M-8': (Keys.Escape, '8'),
    'M-9': (Keys.Escape, '9'),

    'M-Up': (Keys.Escape, Keys.Up),
    'M-Down': (Keys.Escape, Keys.Down, ),
    'M-Left': (Keys.Escape, Keys.Left, ),
    'M-Right': (Keys.Escape, Keys.Right, ),
    'Left': (Keys.Left, ),
    'Right': (Keys.Right, ),
    'Up': (Keys.Up, ),
    'Down': (Keys.Down, ),
    'BSpace': (Keys.Backspace, ),
    'BTab': (Keys.BackTab, ),
    'DC': (Keys.Delete, ),
    'IC': (Keys.Insert, ),
    'End': (Keys.End, ),
    'Enter': (Keys.ControlJ, ),
    'Home': (Keys.Home, ),
    'Escape': (Keys.Escape, ),
    'Tab': (Keys.Tab, ),

    'F1': (Keys.F1, ),
    'F2': (Keys.F2, ),
    'F3': (Keys.F3, ),
    'F4': (Keys.F4, ),
    'F5': (Keys.F5, ),
    'F6': (Keys.F6, ),
    'F7': (Keys.F7, ),
    'F8': (Keys.F8, ),
    'F9': (Keys.F9, ),
    'F10': (Keys.F10, ),
    'F11': (Keys.F11, ),
    'F12': (Keys.F12, ),
    'F13': (Keys.F13, ),
    'F14': (Keys.F14, ),
    'F15': (Keys.F15, ),
    'F16': (Keys.F16, ),
    'F17': (Keys.F17, ),
    'F18': (Keys.F18, ),
    'F19': (Keys.F19, ),
    'F20': (Keys.F20, ),

    'NPage': (Keys.PageDown, ),
    'PageDown': (Keys.PageDown, ),
    'PgDn': (Keys.PageDown, ),
    'PPage': (Keys.PageUp, ),
    'PageUp': (Keys.PageUp, ),
    'PgUp': (Keys.PageUp, ),
}



from __future__ import unicode_literals

__all__ = (
    'wrap_argument',
)


def wrap_argument(text):
    """
    Wrap command argument in quotes and escape when this contains special characters.
    """
    if not any(x in text for x in [' ', '"', "'", '\\']):
        return text
    else:
        return '"%s"' % (text.replace('\\', r'\\').replace('"', r'\"'), )



"""
Improvements on Pyte.
"""
from __future__ import unicode_literals
from pyte.streams import Stream
from pyte.escape import NEL
from pyte import control as ctrl
from collections import defaultdict

from .log import logger

import re

__all__ = (
    'BetterStream',
)


class BetterStream(Stream):
    """
    Extension to the Pyte `Stream` class that also handles "Esc]<num>...BEL"
    sequences. This is used by xterm to set the terminal title.
    """
    escape = Stream.escape.copy()
    escape.update({
        # Call next_line instead of line_feed. We always want to go to the left
        # margin if we receive this, unlike \n, which goes one row down.
        # (Except when LNM has been set.)
        NEL: "next_line",
    })

    def __init__(self, screen):
        super(BetterStream, self).__init__()
        self.listener = screen

        self._validate_screen()

        # Create a regular expression pattern that matches everything what can
        # be considered plain text. This can be used as a very simple lexer
        # that can feed the "plain text part" as one token into the screen.
        special = set([ctrl.ESC, ctrl.CSI, ctrl.NUL, ctrl.DEL]) | set(self.basic)
        self._text_search = re.compile(
            '[^%s]+' % ''.join(re.escape(c) for c in special)).match

        # Start parser.
        self._parser = self._parser_generator()
        self._taking_plain_text = self._parser.send(None)
        self._send = self._parser.send

    def _validate_screen(self):
        """
        Check whether our Screen class has all the required callbacks.
        (We want to verify this statically, before feeding content to the
        screen.)
        """
        for d in [self.basic, self.escape, self.sharp, self.percent, self.csi]:
            for name in d.values():
                assert hasattr(self.listener, name), 'Screen is missing %r' % name

    def feed(self, chars):
        """
        Custom, much more efficient 'feed' function.
        Feed a string of characters to the parser.
        """
        # The original implementation of this function looked like this::
        #
        #     for c in chars:
        #         self._send(c)
        #
        # However, the implementation below does a big optimization. If the
        # parser is possibly expecting a chunk of text (when it is not inside a
        # ESC or CSI escape sequence), then we send that fragment directly to
        # the 'draw' method of the screen.

        # Local copy of functions. (For faster lookups.)
        send = self._send
        taking_plain_text = self._taking_plain_text
        text_search = self._text_search
        draw = self.listener.draw

        # Loop through the chars.
        i = 0
        count = len(chars)

        while i < count:
            # Reading plain text? Don't send characters one by one in the
            # generator, but optimize and send the whole chunk without
            # escapes directly to the listener.
            if taking_plain_text:
                match = text_search(chars, i)
                if match:
                    start, i = match.span()
                    draw(chars[start:i])
                else:
                    taking_plain_text = False

            # The parser expects just one character now. Just send the next one.
            else:
                taking_plain_text = send(chars[i])
                i += 1

        # Remember state for the next 'feed()'.
        self._taking_plain_text = taking_plain_text

    def _parser_generator(self):
        """
        Coroutine that processes VT100 output.

        It's actually a state machine, implemented as a coroutine. So all the
        'state' that we have is stored in local variables.

        This generator is not the most beautiful, but it is as performant as
        possible. When a process generates a lot of output (That is often much
        more than a person would give as input), then this will be the
        bottleneck, because it processes just one character at a time.

        We did many manual optimizations to this function in order to make it
        as efficient as possible. Don't change anything without profiling
        first.
        """
        listener = self.listener

        basic = self.basic
        escape = self.escape
        sharp = self.sharp
        percent = self.percent
        csi = self.csi

        ESC = ctrl.ESC
        CSI = ctrl.CSI
        CTRL_SEQUENCES_ALLOWED_IN_CSI = set([
            ctrl.BEL, ctrl.BS, ctrl.HT, ctrl.LF, ctrl.VT, ctrl.FF, ctrl.CR])

        def create_dispatch_dictionary(source_dict):
            # In order to avoid getting KeyError exceptions below, we make sure
            # that these dictionaries have a dummy handler.
            def dummy(*a, **kw):
                pass

            return defaultdict(
                lambda: dummy,
                dict((event, getattr(listener, attr)) for event, attr in source_dict.items()))

        basic_dispatch = create_dispatch_dictionary(basic)
        sharp_dispatch = create_dispatch_dictionary(sharp)
        percent_dispatch = create_dispatch_dictionary(percent)
        escape_dispatch = create_dispatch_dictionary(escape)
        csi_dispatch = create_dispatch_dictionary(csi)

        while True:
            char = yield True  # (`True` tells the 'send()' function that it
                               # is allowed to send chunks of plain text
                               # directly to the listener, instead of this generator.)

            if char == ESC:  # \x1b
                char = yield

                if char == '[':
                    char = CSI  # Go to CSI.
                else:
                    if char == '#':
                        sharp_dispatch[(yield)]()
                    elif char == '%':
                        percent_dispatch[(yield)]()
                    elif char in '()':
                        listener.set_charset((yield), mode=char)
                    elif char == ']':
                        data = []
                        while True:
                            c = yield
                            if c == '\07':
                                break
                            else:
                                data.append(c)
                        listener.square_close(''.join(data))
                    else:
                        escape_dispatch[char]()
                    continue  # Do not go to CSI.

            if char in basic:  # 'if', not 'elif', because we need to be
                               # able to jump here from Esc[ above in the CSI
                               # section below.
                basic_dispatch[char]()

            elif char == CSI:  # \x9b
                current = ''
                params = []
                private = False

                while True:
                    char = yield
                    if char == '?':
                        private = True
                    elif char in CTRL_SEQUENCES_ALLOWED_IN_CSI:
                        basic_dispatch[char]()
                    elif char in (ctrl.SP, '>'):
                        # Ignore '>' because of 'ESC[>c' (Send device attributes.)
                        pass
                    elif char.isdigit():
                        current += char
                    else:
                        params.append(min(int(current or 0), 9999))

                        if char == ';':
                            current = ''
                        else:
                            try:
                                if private:
                                    csi_dispatch[char](*params, private=True)
                                else:
                                    csi_dispatch[char](*params)
                            except TypeError:
                                # Handler doesn't take params or private attribute.
                                # (Not the cleanest way to handle this, but
                                # it's safe and performant enough.)
                                logger.warning('Dispatch %s failed. params=%s, private=%s',
                                               params, private)
                            break  # Break outside CSI loop.



"""
The child process.
"""
from __future__ import unicode_literals

from prompt_toolkit.document import Document
from prompt_toolkit.eventloop.base import EventLoop
from prompt_toolkit.eventloop.posix_utils import PosixStdinReader
from six.moves import range

from .key_mappings import prompt_toolkit_key_to_vt100_key
from .screen import BetterScreen
from .stream import BetterStream
from .utils import set_terminal_size, pty_make_controlling_tty

import os
import resource
import signal
import sys
import time
import traceback

__all__ = (
    'Process',
)


class Process(object):
    """
    Child process.
    Functionality for parsing the vt100 output (the Pyte screen and stream), as
    well as sending input to the process.

    Usage:

        p = Process(eventloop, ...):
        p.start()

    :param eventloop: Prompt_toolkit eventloop. Used for executing blocking
        stuff in an executor, as well as adding additional readers to the
        eventloop.
    :param invalidate: When the screen content changes, and the renderer needs
        to redraw the output, this callback is called.
    :param exec_func: Callable that is called in the child process. (Usualy,
        this calls execv.)
    :param bell_func: Called when the process does a `bell`.
    :param done_callback: Called when the process terminates.
    :param has_priority: Callable that returns True when this Process should
        get priority in the event loop. (When this pane has the focus.)
        Otherwise output can be delayed.
    """
    def __init__(self, eventloop, invalidate, exec_func, bell_func=None,
                 done_callback=None, has_priority=None):
        assert isinstance(eventloop, EventLoop)
        assert callable(invalidate)
        assert callable(exec_func)
        assert bell_func is None or callable(bell_func)
        assert done_callback is None or callable(done_callback)
        assert has_priority is None or callable(has_priority)

        self.eventloop = eventloop
        self.invalidate = invalidate
        self.exec_func = exec_func
        self.done_callback = done_callback
        self.has_priority = has_priority or (lambda: True)

        self.pid = None
        self.is_terminated = False
        self.suspended = False

        # Create pseudo terminal for this pane.
        self.master, self.slave = os.openpty()

        # Master side -> attached to terminal emulator.
        self._reader = PosixStdinReader(self.master, errors='replace')

        # Create output stream and attach to screen
        self.sx = 0
        self.sy = 0

        self.screen = BetterScreen(self.sx, self.sy,
                                   write_process_input=self.write_input,
                                   bell_func=bell_func)

        self.stream = BetterStream(self.screen)
        self.stream.attach(self.screen)

    def start(self):
        """
        Start the process: fork child.
        """
        self.set_size(120, 24)
        self._start()
        self._connect_reader()
        self._waitpid()

    @classmethod
    def from_command(cls, eventloop, invalidate, command, done_callback,
                     bell_func=None, before_exec_func=None, has_priority=None):
        """
        Create Process from command,
        e.g. command=['python', '-c', 'print("test")']

        :param before_exec_func: Function that is called before `exec` in the process fork.
        """
        assert isinstance(command, list)

        def execv():
            if before_exec_func:
                before_exec_func()

            for p in os.environ['PATH'].split(':'):
                path = os.path.join(p, command[0])
                if os.path.exists(path) and os.access(path, os.X_OK):
                    os.execv(path, command)

        return cls(eventloop, invalidate, execv,
                   bell_func=bell_func, done_callback=done_callback,
                   has_priority=has_priority)

    def _start(self):
        """
        Create fork and start the child process.
        """
        pid = os.fork()

        if pid == 0:
            self._in_child()
        elif pid > 0:
            # In parent.
            os.close(self.slave)
            self.slave = None

            # We wait a very short while, to be sure the child had the time to
            # call _exec. (Otherwise, we are still sharing signal handlers and
            # FDs.) Resizing the pty, when the child is still in our Python
            # code and has the signal handler from prompt_toolkit, but closed
            # the 'fd' for 'call_from_executor', will cause OSError.
            time.sleep(0.1)

            self.pid = pid

    def _waitpid(self):
        """
        Create an executor that waits and handles process termination.
        """
        def wait_for_finished():
            " Wait for PID in executor. "
            os.waitpid(self.pid, 0)
            self.eventloop.call_from_executor(done)

        def done():
            " PID received. Back in the main thread. "
            # Close pty and remove reader.
            os.close(self.master)
            self.eventloop.remove_reader(self.master)
            self.master = None

            # Callback.
            self.is_terminated = True
            self.done_callback()

        self.eventloop.run_in_executor(wait_for_finished)

    def set_size(self, width, height):
        """
        Set terminal size.
        """
        assert isinstance(width, int)
        assert isinstance(height, int)

        if self.master is not None:
            if (self.sx, self.sy) != (width, height):
                set_terminal_size(self.master, height, width)
        self.screen.resize(lines=height, columns=width)

        self.screen.lines = height
        self.screen.columns = width

        self.sx = width
        self.sy = height

    def _in_child(self):
        " Will be executed in the forked child. "
        os.close(self.master)

        # Remove signal handler for SIGWINCH as early as possible.
        # (We don't want this to be triggered when execv has not been called
        # yet.)
        signal.signal(signal.SIGWINCH, 0)

        pty_make_controlling_tty(self.slave)

        # In the fork, set the stdin/out/err to our slave pty.
        os.dup2(self.slave, 0)
        os.dup2(self.slave, 1)
        os.dup2(self.slave, 2)

        # Execute in child.
        try:
            self._close_file_descriptors()
            self.exec_func()
        except Exception:
            traceback.print_exc()
            time.sleep(5)

            os._exit(1)
        os._exit(0)

    def _close_file_descriptors(self):
        # Do not allow child to inherit open file descriptors from parent.
        # (In case that we keep running Python code. We shouldn't close them.
        # because the garbage collector is still active, and he will close them
        # eventually.)
        max_fd = resource.getrlimit(resource.RLIMIT_NOFILE)[-1]

        try:
            os.closerange(3, max_fd)
        except OverflowError:
            # On OS X, max_fd can return very big values, than closerange
            # doesn't understand, e.g. 9223372036854775807. In this case, just
            # use 4096. This is what Linux systems report, and should be
            # sufficient. (I hope...)
            os.closerange(3, 4096)

    def write_input(self, data, paste=False):
        """
        Write user key strokes to the input.

        :param data: (text, not bytes.) The input.
        :param paste: When True, and the process running here understands
            bracketed paste. Send as pasted text.
        """
        # send as bracketed paste?
        if paste and self.screen.bracketed_paste_enabled:
            data = '\x1b[200~' + data + '\x1b[201~'

        self.write_bytes(data.encode('utf-8'))

    def write_bytes(self, data):
        while self.master is not None:
            try:
                os.write(self.master, data)
            except OSError as e:
                # This happens when the window resizes and a SIGWINCH was received.
                # We get 'Error: [Errno 4] Interrupted system call'
                if e.errno == 4:
                    continue
            return

    def write_key(self, key):
        """
        Write prompt_toolkit Key.
        """
        data = prompt_toolkit_key_to_vt100_key(
            key, application_mode=self.screen.in_application_mode)
        self.write_input(data)

    def _connect_reader(self):
        """
        Process stdout output from the process.
        """
        if self.master is not None:
            self.eventloop.add_reader(self.master, self._read)

    def _read(self):
        """
        Read callback, called by the eventloop.
        """
        d = self._reader.read(4096)  # Make sure not to read too much at once. (Otherwise,
                                     # this could block the event loop.)

        if not self._reader.closed:
            def process():
                self.stream.feed(d)
                self.invalidate()

            # Feed directly, if this process has priority. (That is when this
            # pane has the focus in any of the clients.)
            if self.has_priority():
                process()

            # Otherwise, postpone processing until we have CPU time available.
            else:
                if self.master is not None:
                    self.eventloop.remove_reader(self.master)

                def do_asap():
                    " Process output and reconnect to event loop. "
                    process()
                    self._connect_reader()

                # When the event loop is saturated because of CPU, we will
                # postpone this processing max 'x' seconds.

                # '1' seems like a reasonable value, because that way we say
                # that we will process max 1k/1s in case of saturation.
                # That should be enough to prevent the UI from feeling
                # unresponsive.
                timestamp = time.time() + 1

                self.eventloop.call_from_executor(
                    do_asap, _max_postpone_until=timestamp)
        else:
            # End of stream. Remove child.
            self.eventloop.remove_reader(self.master)

    def suspend(self):
        """
        Suspend process. Stop reading stdout. (Called when going into copy mode.)
        """
        self.suspended = True
        self.eventloop.remove_reader(self.master)

    def resume(self):
        """
        Resume from 'suspend'.
        """
        if self.suspended and self.master is not None:
            self._connect_reader()
            self.suspended = False

    def get_cwd(self):
        """
        The current working directory for this process. (Or `None` when
        unknown.)
        """
        return get_cwd_for_pid(self.pid)

    def get_name(self):
        """
        The name for this process. (Or `None` when unknown.)
        """
        # TODO: Maybe cache for short time.
        if self.master is not None:
            return get_name_for_fd(self.master)

    def send_signal(self, signal):
        " Send signal to running process. "
        assert isinstance(signal, int), type(signal)

        if self.pid and not self.is_terminated:
            try:
                os.kill(self.pid, signal)
            except OSError:
                pass  # [Errno 3] No such process.

    def create_copy_document(self):
        """
        Create a Document instance and token list that can be used in copy
        mode.
        """
        data_buffer = self.screen.data_buffer
        text = []
        token_lists = []

        first_row = min(data_buffer.keys())
        last_row = max(data_buffer.keys())

        def token_has_no_background(token):
            try:
                # Token looks like ('C', color, bgcolor, bold, underline, ...)
                return token[2] is None
            except IndexError:
                return True

        for lineno in range(first_row, last_row + 1):
            token_list = []

            row = data_buffer[lineno]
            max_column = max(row.keys()) if row else 0

            # Remove trailing whitespace. (If the background is transparent.)
            row_data = [row[x] for x in range(0, max_column + 1)]

            while (row_data and row_data[-1].char.isspace() and
                   token_has_no_background(row_data[-1].token)):
                row_data.pop()

            # Walk through row.
            char_iter = iter(range(len(row_data)))

            for x in char_iter:
                c = row[x]
                text.append(c.char)
                token_list.append((c.token, c.char))

                # Skip next cell when this is a double width character.
                if c.width == 2:
                    try:
                        next(char_iter)
                    except StopIteration:
                        pass

            token_lists.append(token_list)
            text.append('\n')

        def get_tokens_for_line(lineno):
            try:
                return token_lists[lineno]
            except IndexError:
                return []

        # Calculate cursor position.
        d = Document(text=''.join(text))

        return Document(text=d.text,
                        cursor_position=d.translate_row_col_to_index(
                            row=self.screen.pt_screen.cursor_position.y,
                            col=self.screen.pt_screen.cursor_position.x)), get_tokens_for_line


def get_cwd_for_pid(pid):
    """
    Return the current working directory for a given process ID.
    """
    if sys.platform in ('linux', 'linux2', 'cygwin'):
        try:
            return os.readlink('/proc/%s/cwd' % pid)
        except OSError:
            pass


if sys.platform in ('linux', 'linux2', 'cygwin'):
    def get_name_for_fd(fd):
        """
        Return the process name for a given process ID.
        """
        try:
            pgrp = os.tcgetpgrp(fd)
        except OSError:
            # See: https://github.com/jonathanslenders/pymux/issues/46
            return

        try:
            with open('/proc/%s/cmdline' % pgrp, 'rb') as f:
                return f.read().decode('utf-8', 'ignore').partition('\0')[0]
        except IOError:
            pass
elif sys.platform == 'darwin':
    from .darwin import get_proc_name

    def get_name_for_fd(fd):
        """
        Return the process name for a given process ID.
        """
        try:
            pgrp = os.tcgetpgrp(fd)
        except OSError:
            return

        try:
            return get_proc_name(pgrp)
        except IOError:
            pass
else:
    def get_name_for_fd(fd):
        """
        Return the process name for a given process ID.
        """
        return



"""
Some utilities.
"""
from __future__ import unicode_literals
import array
import fcntl
import getpass
import os
import pwd
import sys
import termios

__all__ = (
    'pty_make_controlling_tty',
    'daemonize',
    'set_terminal_size',
    'nonblocking',
    'get_default_shell',
)


def pty_make_controlling_tty(tty_fd):
    """
    This makes the pseudo-terminal the controlling tty. This should be
    more portable than the pty.fork() function. Specifically, this should
    work on Solaris.

    Thanks to pexpect:
    http://pexpect.sourceforge.net/pexpect.html
    """
    child_name = os.ttyname(tty_fd)

    # Disconnect from controlling tty. Harmless if not already connected.
    try:
        fd = os.open("/dev/tty", os.O_RDWR | os.O_NOCTTY)
        if fd >= 0:
            os.close(fd)
    # which exception, shouldnt' we catch explicitly .. ?
    except:
        # Already disconnected. This happens if running inside cron.
        pass

    os.setsid()

    # Verify we are disconnected from controlling tty
    # by attempting to open it again.
    try:
        fd = os.open("/dev/tty", os.O_RDWR | os.O_NOCTTY)
        if fd >= 0:
            os.close(fd)
            raise Exception('Failed to disconnect from controlling '
                            'tty. It is still possible to open /dev/tty.')
    # which exception, shouldnt' we catch explicitly .. ?
    except:
        # Good! We are disconnected from a controlling tty.
        pass

    # Verify we can open child pty.
    fd = os.open(child_name, os.O_RDWR)
    if fd < 0:
        raise Exception("Could not open child pty, " + child_name)
    else:
        os.close(fd)

    # Verify we now have a controlling tty.
    if os.name != 'posix':
        # Skip this on BSD-like systems since it will break.
        fd = os.open("/dev/tty", os.O_WRONLY)
        if fd < 0:
            raise Exception("Could not open controlling tty, /dev/tty")
        else:
            os.close(fd)


def daemonize(stdin='/dev/null', stdout='/dev/null', stderr='/dev/null'):
    """
    Double fork-trick. For starting a posix daemon.

    This forks the current process into a daemon. The stdin, stdout, and stderr
    arguments are file names that will be opened and be used to replace the
    standard file descriptors in sys.stdin, sys.stdout, and sys.stderr. These
    arguments are optional and default to /dev/null. Note that stderr is opened
    unbuffered, so if it shares a file with stdout then interleaved output may
    not appear in the order that you expect.

    Thanks to:
    http://code.activestate.com/recipes/66012-fork-a-daemon-process-on-unix/
    """
    # Do first fork.
    try:
        pid = os.fork()
        if pid > 0:
            os.waitpid(pid, 0)
            return 0  # Return 0 from first parent.
    except OSError as e:
        sys.stderr.write("fork #1 failed: (%d) %s\n" % (e.errno, e.strerror))
        sys.exit(1)

    # Decouple from parent environment.
    os.chdir("/")
    os.umask(0)
    os.setsid()

    # Do second fork.
    try:
        pid = os.fork()
        if pid > 0:
            sys.exit(0)  # Exit second parent.
    except OSError as e:
        sys.stderr.write("fork #2 failed: (%d) %s\n" % (e.errno, e.strerror))
        sys.exit(1)

    # Now I am a daemon!

    # Redirect standard file descriptors.

        # NOTE: For debugging, you meight want to take these instead of /dev/null.
    # so = open('/tmp/log2', 'ab+')
    # se = open('/tmp/log2', 'ab+', 0)

    si = open(stdin, 'rb')
    so = open(stdout, 'ab+')
    se = open(stderr, 'ab+', 0)
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())

    # Return 1 from daemon.
    return 1


def set_terminal_size(stdout_fileno, rows, cols):
    """
    Set terminal size.

    (This is also mainly for internal use. Setting the terminal size
    automatically happens when the window resizes. However, sometimes the
    process that created a pseudo terminal, and the process that's attached to
    the output window are not the same, e.g. in case of a telnet connection, or
    unix domain socket, and then we have to sync the sizes by hand.)
    """
    # Buffer for the C call
    # (The first parameter of 'array.array' needs to be 'str' on both Python 2
    # and Python 3.)
    buf = array.array(str('h'), [rows, cols, 0, 0])

    # Do: TIOCSWINSZ (Set)
    fcntl.ioctl(stdout_fileno, termios.TIOCSWINSZ, buf)


class nonblocking(object):
    """
    Make fd non blocking.
    """
    def __init__(self, fd):
        self.fd = fd

    def __enter__(self):
        self.orig_fl = fcntl.fcntl(self.fd, fcntl.F_GETFL)
        fcntl.fcntl(self.fd, fcntl.F_SETFL, self.orig_fl | os.O_NONBLOCK)

    def __exit__(self, *args):
        fcntl.fcntl(self.fd, fcntl.F_SETFL, self.orig_fl)


def get_default_shell():
    """
    return the path to the default shell for the current user.
    """
    if 'SHELL' in os.environ:
        return os.environ['SHELL']
    else:
        username = getpass.getuser()
        shell = pwd.getpwnam(username).pw_shell
        return shell



from __future__ import unicode_literals

from prompt_toolkit.eventloop.base import INPUT_TIMEOUT
from prompt_toolkit.eventloop.posix import call_on_sigwinch
from prompt_toolkit.eventloop.select import select_fds
from prompt_toolkit.eventloop.posix_utils import PosixStdinReader
from prompt_toolkit.terminal.vt100_input import raw_mode, cooked_mode
from prompt_toolkit.terminal.vt100_output import _get_size, Vt100_Output

from pymux.utils import nonblocking

import getpass
import glob
import json
import os
import signal
import socket
import sys
import tempfile


__all__ = (
    'Client',
    'list_clients',
)


class Client(object):
    def __init__(self, socket_name):
        self.socket_name = socket_name
        self._mode_context_managers = []

        # Connect to socket.
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.socket.connect(socket_name)
        self.socket.setblocking(1)

        # Input reader.
        #     Some terminals, like lxterminal send non UTF-8 input sequences,
        #     even when the input encoding is supposed to be UTF-8. This
        #     happens in the case of mouse clicks in the right area of a wide
        #     terminal. Apparently, these are some binary blobs in between the
        #     UTF-8 input.)
        #     We should not replace these, because this would break the
        #     decoding otherwise. (Also don't pass errors='ignore', because
        #     that doesn't work for parsing mouse input escape sequences, which
        #     consist of a fixed number of bytes.)
        self._stdin_reader = PosixStdinReader(sys.stdin.fileno(), errors='replace')

    def run_command(self, command, pane_id=None):
        """
        Ask the server to run this command.

        :param pane_id: Optional identifier of the current pane.
        """
        self._send_packet({
            'cmd': 'run-command',
            'data': command,
            'pane_id': pane_id
        })

    def attach(self, detach_other_clients=False, ansi_colors_only=False, true_color=False):
        """
        Attach client user interface.
        """
        assert isinstance(detach_other_clients, bool)
        assert isinstance(ansi_colors_only, bool)
        assert isinstance(true_color, bool)

        self._send_size()
        self._send_packet({
            'cmd': 'start-gui',
            'detach-others': detach_other_clients,
            'ansi-colors-only': ansi_colors_only,
            'true-color': true_color,
            'term': os.environ.get('TERM', ''),
            'data': ''
        })

        with raw_mode(sys.stdin.fileno()):
            data_buffer = b''

            stdin_fd = sys.stdin.fileno()
            socket_fd = self.socket.fileno()
            current_timeout = INPUT_TIMEOUT  # Timeout, used to flush escape sequences.

            with call_on_sigwinch(self._send_size):
                while True:
                    r = select_fds([stdin_fd, socket_fd], current_timeout)

                    if socket_fd in r:
                        # Received packet from server.
                        data = self.socket.recv(1024)

                        if data == b'':
                            # End of file. Connection closed.
                            # Reset terminal
                            o = Vt100_Output.from_pty(sys.stdout)
                            o.quit_alternate_screen()
                            o.disable_mouse_support()
                            o.disable_bracketed_paste()
                            o.reset_attributes()
                            o.flush()
                            return
                        else:
                            data_buffer += data

                            while b'\0' in data_buffer:
                                pos = data_buffer.index(b'\0')
                                self._process(data_buffer[:pos])
                                data_buffer = data_buffer[pos + 1:]

                    elif stdin_fd in r:
                        # Got user input.
                        self._process_stdin()
                        current_timeout = INPUT_TIMEOUT

                    else:
                        # Timeout. (Tell the server to flush the vt100 Escape.)
                        self._send_packet({'cmd': 'flush-input'})
                        current_timeout = None

    def _process(self, data_buffer):
        """
        Handle incoming packet from server.
        """
        packet = json.loads(data_buffer.decode('utf-8'))

        if packet['cmd'] == 'out':
            # Call os.write manually. In Python2.6, sys.stdout.write doesn't use UTF-8.
            os.write(sys.stdout.fileno(), packet['data'].encode('utf-8'))

        elif packet['cmd'] == 'suspend':
            # Suspend client process to background.
            if hasattr(signal, 'SIGTSTP'):
                os.kill(os.getpid(), signal.SIGTSTP)

        elif packet['cmd'] == 'mode':
            # Set terminal to raw/cooked.
            action = packet['data']

            if action == 'raw':
                cm = raw_mode(sys.stdin.fileno())
                cm.__enter__()
                self._mode_context_managers.append(cm)

            elif action == 'cooked':
                cm = cooked_mode(sys.stdin.fileno())
                cm.__enter__()
                self._mode_context_managers.append(cm)

            elif action == 'restore' and self._mode_context_managers:
                cm = self._mode_context_managers.pop()
                cm.__exit__()

    def _process_stdin(self):
        """
        Received data on stdin. Read and send to server.
        """
        with nonblocking(sys.stdin.fileno()):
            data = self._stdin_reader.read()

        # Send input in chunks of 4k.
        step = 4056
        for i in range(0, len(data), step):
            self._send_packet({
                'cmd': 'in',
                'data': data[i:i + step],
            })

    def _send_packet(self, data):
        " Send to server. "
        data = json.dumps(data).encode('utf-8')

        # Be sure that our socket is blocking, otherwise, the send() call could
        # raise `BlockingIOError` if the buffer is full.
        self.socket.setblocking(1)

        self.socket.send(data + b'\0')

    def _send_size(self):
        " Report terminal size to server. "
        rows, cols = _get_size(sys.stdout.fileno())
        self._send_packet({
            'cmd': 'size',
            'data': [rows, cols]
        })


def list_clients():
    """
    List all the servers that are running.
    """
    p = '%s/pymux.sock.%s.*' % (tempfile.gettempdir(), getpass.getuser())
    for path in glob.glob(p):
        try:
            yield Client(path)
        except socket.error:
            pass



"""
Arrangement of panes.

Don't confuse with the prompt_toolkit VSplit/HSplit classes. This is a higher
level abstraction of the Pymux window layout.

An arrangement consists of a list of windows. And a window has a list of panes,
arranged by ordering them in HSplit/VSplit instances.
"""
from __future__ import unicode_literals
from .process import Process

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.interface import CommandLineInterface
from prompt_toolkit.search_state import SearchState
from prompt_toolkit.token import Token

import math
import os
import weakref
import six

__all__ = (
    'LayoutTypes',
    'Pane',
    'HSplit',
    'VSplit',
    'Window',
    'Arrangement',
)


class LayoutTypes:
    # The values are in lowercase with dashes, because that is what users can
    # use at the command line.
    EVEN_HORIZONTAL = 'even-horizontal'
    EVEN_VERTICAL = 'even-vertical'
    MAIN_HORIZONTAL = 'main-horizontal'
    MAIN_VERTICAL = 'main-vertical'
    TILED = 'tiled'

    _ALL = [EVEN_HORIZONTAL, EVEN_VERTICAL, MAIN_HORIZONTAL, MAIN_VERTICAL, TILED]


class Pane(object):
    """
    One pane, containing one process and a search buffer for going into copy
    mode or displaying the help.
    """
    _pane_counter = 1000  # Start at 1000, to be sure to not confuse this with pane indexes.

    def __init__(self, process):
        assert isinstance(process, Process)

        self.process = process
        self.chosen_name = None

        # Displayed the clock instead of this pane content.
        self.clock_mode = False

        # Give unique ID.
        Pane._pane_counter += 1
        self.pane_id = Pane._pane_counter

        # Prompt_toolkit buffer, for displaying scrollable text.
        # (In copy mode, or help mode.)
        # Note: Because the scroll_buffer can only contain text, we also use the
        #       get_tokens_for_line, that returns the token list with color
        #       information for each line.
        self.scroll_buffer = Buffer(read_only=True)
        self.copy_get_tokens_for_line = lambda lineno: []
        self.display_scroll_buffer = False
        self.scroll_buffer_title = ''

        # Search buffer, for use in copy mode. (Each pane gets its own search buffer.)
        self.search_buffer = Buffer()
        self.is_searching = False
        self.search_state = SearchState(ignore_case=False)

    @property
    def name(self):
        """
        The name for the window as displayed in the title bar and status bar.
        """
        # Name, explicitely set for the pane.
        if self.chosen_name:
            return self.chosen_name
        else:
            # Name from the process running inside the pane.
            name = self.process.get_name()
            if name:
                return os.path.basename(name)

        return ''

    def enter_copy_mode(self):
        """
        Suspend the process, and copy the screen content to the `scroll_buffer`.
        That way the user can search through the history and copy/paste.
        """
        document, get_tokens_for_line = self.process.create_copy_document()
        self._enter_scroll_buffer('Copy', document, get_tokens_for_line)

    def display_text(self, text, title=''):
        """
        Display the given text in the scroll buffer.
        """
        document = Document(text, 0)

        def get_tokens_for_line(lineno):
            return [(Token, document.lines[lineno])]

        self._enter_scroll_buffer(
            title,
            document=document,
            get_tokens_for_line=get_tokens_for_line)

    def _enter_scroll_buffer(self, title, document, get_tokens_for_line):
        # Suspend child process.
        self.process.suspend()

        self.scroll_buffer.set_document(document, bypass_readonly=True)
        self.copy_get_tokens_for_line = get_tokens_for_line
        self.display_scroll_buffer = True
        self.scroll_buffer_title = title

        # Reset search state.
        self.search_state = SearchState(ignore_case=False)

    def exit_scroll_buffer(self):
        """
        Exit scroll buffer. (Exits help or copy mode.)
        """
        self.process.resume()
        self.display_scroll_buffer = False


class _WeightsDictionary(weakref.WeakKeyDictionary):
    """
    Dictionary for the weights: weak keys, but defaults to 1.

    (Weights are used to represent the proportion of pane sizes in
    HSplit/VSplit lists.)

    This dictionary maps the child (another HSplit/VSplit or Pane), to the
    size. (Integer.)
    """
    def __getitem__(self, key):
        try:
            # (Don't use 'super' here. This is a classobj in Python2.)
            return weakref.WeakKeyDictionary.__getitem__(self, key)
        except KeyError:
            return 1


class _Split(list):
    """
    Base class for horizontal and vertical splits. (This is a higher level
    split than prompt_toolkit.layout.HSplit.)
    """
    def __init__(self, *a, **kw):
        list.__init__(self, *a, **kw)

        # Mapping children to its weight.
        self.weights = _WeightsDictionary()

    def __hash__(self):
        # Required in order to add HSplit/VSplit to the weights dict. "
        return id(self)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, list.__repr__(self))


class HSplit(_Split):
    """ Horizontal split. """


class VSplit(_Split):
    """ Horizontal split. """


class Window(object):
    """
    Pymux window.
    """
    _window_counter = 1000  # Start here, to avoid confusion with window index.

    def __init__(self, index=0):
        self.index = index
        self.root = HSplit()
        self._active_pane = None
        self._prev_active_pane = None
        self.chosen_name = None
        self.previous_selected_layout = None

        #: When true, the current pane is zoomed in.
        self.zoom = False

        #: When True, send input to all panes simultaniously.
        self.synchronize_panes = False

        # Give unique ID.
        Window._window_counter += 1
        self.window_id = Window._window_counter

    def invalidation_hash(self):
        """
        Return a hash (string) that can be used to determine when the layout
        has to be rebuild.
        """
        def _hash_for_split(split):
            result = []
            for item in split:
                if isinstance(item, (VSplit, HSplit)):
                    result.append(_hash_for_split(item))
                elif isinstance(item, Pane):
                    result.append('p%s' % item.pane_id)

            if isinstance(split, HSplit):
                return 'HSplit(%s)' % (','.join(result))
            else:
                return 'VSplit(%s)' % (','.join(result))

        return '<window_id=%s,zoom=%s,children=%s>' % (
            self.window_id, self.zoom, _hash_for_split(self.root))

    @property
    def active_pane(self):
        """
        The current active :class:`.Pane`.
        """
        return self._active_pane

    @active_pane.setter
    def active_pane(self, value):
        assert isinstance(value, Pane)

        # Remember previous active pane.
        if self._active_pane:
            self._prev_active_pane = weakref.ref(self._active_pane)

        self.zoom = False
        self._active_pane = value

    @property
    def previous_active_pane(self):
        """
        The previous active :class:`.Pane` or `None` if unknown.
        """
        p = self._prev_active_pane and self._prev_active_pane()

        # Only return when this pane actually still exists in the current
        # window.
        if p and p in self.panes:
            return p

    @property
    def name(self):
        """
        The name for this window as it should be displayed in the status bar.
        """
        # Name, explicitely set for the window.
        if self.chosen_name:
            return self.chosen_name
        else:
            pane = self.active_pane
            if pane:
                return pane.name

        return ''

    def add_pane(self, pane, vsplit=False):
        """
        Add another pane to this Window.
        """
        assert isinstance(pane, Pane)
        assert isinstance(vsplit, bool)

        split_cls = VSplit if vsplit else HSplit

        if self.active_pane is None:
            self.root.append(pane)
        else:
            parent = self._get_parent(self.active_pane)
            same_direction = isinstance(parent, split_cls)

            index = parent.index(self.active_pane)

            if same_direction:
                parent.insert(index + 1, pane)
            else:
                new_split = split_cls([self.active_pane, pane])
                parent[index] = new_split

                # Give the newly created split the same weight as the original
                # pane that was at this position.
                parent.weights[new_split] = parent.weights[self.active_pane]

        self.active_pane = pane
        self.zoom = False

    def remove_pane(self, pane):
        """
        Remove pane from this Window.
        """
        assert isinstance(pane, Pane)

        if pane in self.panes:
            # When this pane was focused, switch to previous active or next in order.
            if pane == self.active_pane:
                if self.previous_active_pane:
                    self.active_pane = self.previous_active_pane
                else:
                    self.focus_next()

            # Remove from the parent. When the parent becomes empty, remove the
            # parent itself recursively.
            p = self._get_parent(pane)
            p.remove(pane)

            while len(p) == 0 and p != self.root:
                p2 = self._get_parent(p)
                p2.remove(p)
                p = p2

            # When the parent has only one item left, collapse into its parent.
            while len(p) == 1 and p != self.root:
                p2 = self._get_parent(p)
                p2.weights[p[0]] = p2.weights[p]  # Keep dimensions.
                i = p2.index(p)
                p2[i] = p[0]
                p = p2

    @property
    def panes(self):
        " List with all panes from this Window. "
        result = []

        for s in self.splits:
            for item in s:
                if isinstance(item, Pane):
                    result.append(item)

        return result

    @property
    def splits(self):
        " Return a list with all HSplit/VSplit instances. "
        result = []

        def collect(split):
            result.append(split)

            for item in split:
                if isinstance(item, (HSplit, VSplit)):
                    collect(item)

        collect(self.root)
        return result

    def _get_parent(self, item):
        " The HSplit/VSplit that contains the active pane. "
        for s in self.splits:
            if item in s:
                return s

    @property
    def has_panes(self):
        " True when this window contains at least one pane. "
        return len(self.panes) > 0

    @property
    def active_process(self):
        " Return `Process` that should receive user input. "
        p = self.active_pane

        if p is not None:
            return p.process

    def focus_next(self, count=1):
        " Focus the next pane. "
        panes = self.panes
        self.active_pane = panes[(panes.index(self.active_pane) + count) % len(panes)]

    def focus_previous(self):
        " Focus the previous pane. "
        self.focus_next(count=-1)

    def rotate(self, count=1, with_pane_before_only=False, with_pane_after_only=False):
        """
        Rotate panes.
        When `with_pane_before_only` or `with_pane_after_only` is True, only rotate
        with the pane before/after the active pane.
        """
        # Create (split, index, pane, weight) tuples.
        items = []
        current_pane_index = None

        for s in self.splits:
            for index, item in enumerate(s):
                if isinstance(item, Pane):
                    items.append((s, index, item, s.weights[item]))
                    if item == self.active_pane:
                        current_pane_index = len(items) - 1

        # Only before after? Reduce list of panes.
        if with_pane_before_only:
            items = items[current_pane_index - 1:current_pane_index + 1]

        elif with_pane_after_only:
            items = items[current_pane_index:current_pane_index + 2]

        # Rotate positions.
        for i, triple in enumerate(items):
            split, index, pane, weight = triple

            new_item = items[(i + count) % len(items)][2]

            split[index] = new_item
            split.weights[new_item] = weight

    def select_layout(self, layout_type):
        """
        Select one of the predefined layouts.
        """
        assert layout_type in LayoutTypes._ALL

        # When there is only one pane, always choose EVEN_HORIZONTAL,
        # Otherwise, we create VSplit/HSplit instances with an empty list of
        # children.
        if len(self.panes) == 1:
            layout_type = LayoutTypes.EVEN_HORIZONTAL

        # even-horizontal.
        if layout_type == LayoutTypes.EVEN_HORIZONTAL:
            self.root = HSplit(self.panes)

        # even-vertical.
        elif layout_type == LayoutTypes.EVEN_VERTICAL:
            self.root = VSplit(self.panes)

        # main-horizontal.
        elif layout_type == LayoutTypes.MAIN_HORIZONTAL:
            self.root = HSplit([
                self.active_pane,
                VSplit([p for p in self.panes if p != self.active_pane])
            ])

        # main-vertical.
        elif layout_type == LayoutTypes.MAIN_VERTICAL:
            self.root = VSplit([
                self.active_pane,
                HSplit([p for p in self.panes if p != self.active_pane])
            ])

        # tiled.
        elif layout_type == LayoutTypes.TILED:
            panes = self.panes
            column_count = math.ceil(len(panes) ** .5)

            rows = HSplit()
            current_row = VSplit()

            for p in panes:
                current_row.append(p)

                if len(current_row) >= column_count:
                    rows.append(current_row)
                    current_row = VSplit()
            if current_row:
                rows.append(current_row)

            self.root = rows

        self.previous_selected_layout = layout_type

    def select_next_layout(self, count=1):
        """
        Select next layout. (Cycle through predefined layouts.)
        """
        # List of all layouts. (When we have just two panes, only toggle
        # between horizontal/vertical.)
        if len(self.panes) == 2:
            all_layouts = [LayoutTypes.EVEN_HORIZONTAL, LayoutTypes.EVEN_VERTICAL]
        else:
            all_layouts = LayoutTypes._ALL

        # Get index of current layout.
        layout = self.previous_selected_layout or LayoutTypes._ALL[-1]
        try:
            index = all_layouts.index(layout)
        except ValueError:
            index = 0

        # Switch to new layout.
        new_layout = all_layouts[(index + count) % len(all_layouts)]
        self.select_layout(new_layout)

    def select_previous_layout(self):
        self.select_next_layout(count=-1)

    def change_size_for_active_pane(self, up=0, right=0, down=0, left=0):
        """
        Increase the size of the current pane in any of the four directions.
        """
        child = self.active_pane
        self.change_size_for_pane(child, up=up, right=right, down=down, left=left)

    def change_size_for_pane(self, pane, up=0, right=0, down=0, left=0):
        """
        Increase the size of the current pane in any of the four directions.
        Positive values indicate an increase, negative values a decrease.
        """
        assert isinstance(pane, Pane)

        def find_split_and_child(split_cls, is_before):
            " Find the split for which we will have to update the weights. "
            child = pane
            split = self._get_parent(child)

            def found():
                return isinstance(split, split_cls) and (
                    not is_before or split.index(child) > 0) and (
                        is_before or split.index(child) < len(split) - 1)

            while split and not found():
                child = split
                split = self._get_parent(child)

            return split, child  # split can be None!

        def handle_side(split_cls, is_before, amount, trying_other_side=False):
            " Increase weights on one side. (top/left/right/bottom). "
            if amount:
                split, child = find_split_and_child(split_cls, is_before)

                if split:
                    # Find neighbour.
                    neighbour_index = split.index(child) + (-1 if is_before else 1)
                    neighbour_child = split[neighbour_index]

                    # Increase/decrease weights.
                    split.weights[child] += amount
                    split.weights[neighbour_child] -= amount

                    # Ensure that all weights are at least one.
                    for k, value in split.weights.items():
                        if value < 1:
                            split.weights[k] = 1

                else:
                    # When no split has been found where we can move in this
                    # direction, try to move the other side instead using a
                    # negative amount. This happens when we run "resize-pane -R 4"
                    # inside the pane that is completely on the right. In that
                    # case it's logical to move the left border to the right
                    # instead.
                    if not trying_other_side:
                        handle_side(split_cls, not is_before, -amount,
                                    trying_other_side=True)

        handle_side(VSplit, True, left)
        handle_side(VSplit, False, right)
        handle_side(HSplit, True, up)
        handle_side(HSplit, False, down)

    def get_pane_index(self, pane):
        " Return the index of the given pane. ValueError if not found. "
        assert isinstance(pane, Pane)
        return self.panes.index(pane)


class Arrangement(object):
    """
    Arrangement class for one Pymux session.
    This contains the list of windows and the layout of the panes for each
    window. All the clients share the same Arrangement instance, but they can
    have different windows active.
    """
    def __init__(self):
        self.windows = []
        self.base_index = 0

        self._active_window_for_cli = weakref.WeakKeyDictionary()
        self._prev_active_window_for_cli = weakref.WeakKeyDictionary()

        # The active window of the last CLI. Used as default when a new session
        # is attached.
        self._last_active_window = None

    def pane_has_priority(self, pane):
        """
        Return True when this Pane sohuld get priority in the output processing.
        This is true for panes that have the focus in any of the visible windows.
        """
        windows = set(self._active_window_for_cli.values())

        for w in windows:
            if w.active_pane == pane:
                return True

        return False

    def invalidation_hash(self, cli):
        """
        When this changes, the layout needs to be rebuild.
        """
        w = self.get_active_window(cli)
        return w.invalidation_hash()

    def get_active_window(self, cli):
        """
        The current active :class:`.Window`.
        """
        assert isinstance(cli, CommandLineInterface)

        try:
            return self._active_window_for_cli[cli]
        except KeyError:
            self._active_window_for_cli[cli] = self._last_active_window or self.windows[0]
            return self.windows[0]

    def set_active_window(self, cli, window):
        assert isinstance(cli, CommandLineInterface)
        assert isinstance(window, Window)

        previous = self.get_active_window(cli)
        self._prev_active_window_for_cli[cli] = previous
        self._active_window_for_cli[cli] = window
        self._last_active_window = window

    def set_active_window_from_pane_id(self, cli, pane_id):
        """
        Make the window with this pane ID the active Window.
        """
        assert isinstance(cli, CommandLineInterface)
        assert isinstance(pane_id, int)

        for w in self.windows:
            for p in w.panes:
                if p.pane_id == pane_id:
                    self.set_active_window(cli, w)

    def get_previous_active_window(self, cli):
        " The previous active Window or None if unknown. "
        assert isinstance(cli, CommandLineInterface)

        try:
            return self._prev_active_window_for_cli[cli]
        except KeyError:
            return None

    def get_window_by_index(self, index):
        " Return the Window with this index or None if not found. "
        for w in self.windows:
            if w.index == index:
                return w

    def create_window(self, cli, pane, name=None, set_active=True):
        """
        Create a new window that contains just this pane.

        :param cli: If been given, this window will be focussed for that client.
        :param pane: The :class:`.Pane` instance to put in the new window.
        :param name: If given, name for the new window.
        :param set_active: When True, focus the new window.
        """
        assert isinstance(pane, Pane)
        assert cli is None or isinstance(cli, CommandLineInterface)
        assert name is None or isinstance(name, six.text_type)

        # Take the first available index.
        taken_indexes = [w.index for w in self.windows]

        index = self.base_index
        while index in taken_indexes:
            index += 1

        # Create new window and add it.
        w = Window(index)
        w.add_pane(pane)
        self.windows.append(w)

        # Sort windows by index.
        self.windows = sorted(self.windows, key=lambda w: w.index)

        if cli is not None and set_active:
            self.set_active_window(cli, w)

        if name is not None:
            w.chosen_name = name

        assert w.active_pane == pane
        assert w._get_parent(pane)

    def move_window(self, window, new_index):
        """
        Move window to a new index.
        """
        assert isinstance(window, Window)
        assert isinstance(new_index, int)

        window.index = new_index

        # Sort windows by index.
        self.windows = sorted(self.windows, key=lambda w: w.index)

    def get_active_pane(self, cli):
        """
        The current :class:`.Pane` from the current window.
        """
        assert isinstance(cli, CommandLineInterface)

        w = self.get_active_window(cli)
        if w is not None:
            return w.active_pane

    def remove_pane(self, pane):
        """
        Remove a :class:`.Pane`. (Look in all windows.)
        """
        assert isinstance(pane, Pane)

        for w in self.windows:
            w.remove_pane(pane)

            # No panes left in this window?
            if not w.has_panes:
                # Focus next.
                for cli, active_w in self._active_window_for_cli.items():
                    if w == active_w:
                        self.focus_next_window(cli)

                self.windows.remove(w)

    def focus_previous_window(self, cli):
        assert isinstance(cli, CommandLineInterface)

        w = self.get_active_window(cli)

        self.set_active_window(cli, self.windows[
            (self.windows.index(w) - 1) % len(self.windows)])

    def focus_next_window(self, cli):
        assert isinstance(cli, CommandLineInterface)

        w = self.get_active_window(cli)

        self.set_active_window(cli, self.windows[
            (self.windows.index(w) + 1) % len(self.windows)])

    def break_pane(self, cli, set_active=True):
        """
        When the current window has multiple panes, remove the pane from this
        window and put it in a new window.

        :param set_active: When True, focus the new window.
        """
        assert isinstance(cli, CommandLineInterface)

        w = self.get_active_window(cli)

        if len(w.panes) > 1:
            pane = w.active_pane
            self.get_active_window(cli).remove_pane(pane)
            self.create_window(cli, pane, set_active=set_active)

    def rotate_window(self, cli, count=1):
        " Rotate the panes in the active window. "
        assert isinstance(cli, CommandLineInterface)

        w = self.get_active_window(cli)
        w.rotate(count=count)

    @property
    def has_panes(self):
        " True when any of the windows has a :class:`.Pane`. "
        for w in self.windows:
            if w.has_panes:
                return True
        return False



from __future__ import unicode_literals

__all__ = (
    'COMMAND',
    'PROMPT',
)


#: Name of the command buffer.
COMMAND = 'COMMAND'

#: Name of the input for a "command-prompt" command.
PROMPT = 'PROMPT'



"""
Aliases for all commands.
(On purpose kept compatible with tmux.)
"""
from __future__ import unicode_literals

__all__ = (
    'ALIASES',
)


ALIASES = {
    'bind': 'bind-key',
    'breakp': 'break-pane',
    'clearhist': 'clear-history',
    'confirm': 'confirm-before',
    'detach': 'detach-client',
    'display': 'display-message',
    'displayp': 'display-panes',
    'killp': 'kill-pane',
    'killw': 'kill-window',
    'last': 'last-window',
    'lastp': 'last-pane',
    'lextl': 'next-layout',
    'lsk': 'list-keys',
    'lsp': 'list-panes',
    'movew': 'move-window',
    'neww': 'new-window',
    'next': 'next-window',
    'pasteb': 'paste-buffer',
    'prev': 'previous-window',
    'prevl': 'previous-layout',
    'rename': 'rename-session',
    'renamew': 'rename-window',
    'resizep': 'resize-pane',
    'rotatew': 'rotate-window',
    'selectl': 'select-layout',
    'selectp': 'select-pane',
    'selectw': 'select-window',
    'send': 'send-keys',
    'set': 'set-option',
    'setw': 'set-window-option',
    'source': 'source-file',
    'splitw': 'split-window',
    'suspendc': 'suspend-client',
    'swapp': 'swap-pane',
    'unbind': 'unbind-key',
}



#!/usr/bin/env python
"""
pymux: Pure Python terminal multiplexer.
Usage:
    pymux [(standalone|start-server|attach)] [-d]
          [--truecolor] [--ansicolor] [(-S <socket>)] [(-f <file>)]
          [(--log <logfile>)]
          [--] [<command>]
    pymux list-sessions
    pymux -h | --help
    pymux <command>

Options:
    standalone   : Run as a standalone process. (for debugging, detaching is
                   not possible.
    start-server : Run a server daemon that can be attached later on.
    attach       : Attach to a running session.

    -f           : Path to configuration file. By default: '~/.pymux.conf'.
    -S           : Unix socket path.
    -d           : Detach all other clients, when attaching.
    --log        : Logfile.
    --truecolor  : Render true color (24 bit) instead of 256 colors.
                   (Each client can set this separately.)
"""
from __future__ import unicode_literals, absolute_import

from pymux.main import Pymux
from pymux.client import Client, list_clients
from pymux.utils import daemonize

import docopt
import getpass
import logging
import os
import sys
import tempfile

__all__ = (
    'run',
)


def run():
    a = docopt.docopt(__doc__)
    socket_name = a['<socket>'] or os.environ.get('PYMUX')
    socket_name_from_env = not a['<socket>'] and os.environ.get('PYMUX')
    filename = a['<file>']
    command = a['<command>']
    true_color = a['--truecolor']
    ansi_colors_only = a['--ansicolor'] or \
        bool(os.environ.get('PROMPT_TOOLKIT_ANSI_COLORS_ONLY', False))

    # Parse pane_id from socket_name. It looks like "socket_name,pane_id".
    if socket_name and ',' in socket_name:
        socket_name, pane_id = socket_name.rsplit(',', 1)
    else:
        pane_id = None

    # Expand socket name. (Make it possible to just accept numbers.)
    if socket_name and socket_name.isdigit():
        socket_name = '%s/pymux.sock.%s.%s' % (
            tempfile.gettempdir(), getpass.getuser(), socket_name)

    # Configuration filename.
    default_config = os.path.abspath(os.path.expanduser('~/.pymux.conf'))
    if not filename and os.path.exists(default_config):
        filename = default_config

    if filename:
        filename = os.path.abspath(os.path.expanduser(filename))

    # Create 'Pymux'.
    mux = Pymux(source_file=filename, startup_command=command)

    # Setup logging.
    if a['<logfile>']:
        logging.basicConfig(filename=a['<logfile>'], level=logging.DEBUG)

    if a['standalone']:
        mux.run_standalone(true_color=true_color, ansi_colors_only=ansi_colors_only)

    elif a['list-sessions'] or a['<command>'] in ('ls', 'list-sessions'):
        for c in list_clients():
            print(c.socket_name)

    elif a['start-server']:
        if socket_name_from_env:
            _socket_from_env_warning()
            sys.exit(1)

        # Log to stdout.
        logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

        # Run server.
        socket_name = mux.listen_on_socket()
        try:
            mux.run_server()
        except KeyboardInterrupt:
            sys.exit(1)

    elif a['attach']:
        if socket_name_from_env:
            _socket_from_env_warning()
            sys.exit(1)

        detach_other_clients = a['-d']

        if socket_name:
            Client(socket_name).attach(
                detach_other_clients=detach_other_clients,
                true_color=true_color,
                ansi_colors_only=ansi_colors_only)
        else:
            # Connect to the first server.
            for c in list_clients():
                c.attach(detach_other_clients=detach_other_clients,
                         true_color=true_color,
                         ansi_colors_only=ansi_colors_only)
                break
            else:  # Nobreak.
                print('No pymux instance found.')
                sys.exit(1)

    elif a['<command>'] and socket_name:
        Client(socket_name).run_command(a['<command>'], pane_id)

    elif not socket_name:
        # Run client/server combination.
        socket_name = mux.listen_on_socket(socket_name)
        pid = daemonize()

        if pid > 0:
            # Create window. It is important that this happens in the daemon,
            # because the parent of the process running inside should be this
            # daemon. (Otherwise the `waitpid` call won't work.)
            mux.run_server()
        else:
            Client(socket_name).attach(
                true_color=true_color, ansi_colors_only=ansi_colors_only)

    else:
        if socket_name_from_env:
            _socket_from_env_warning()
            sys.exit(1)
        else:
            print('Invalid command.')
            sys.exit(1)


def _socket_from_env_warning():
    print('Please be careful nesting pymux sessions.')
    print('Unset PYMUX environment variable first.')


if __name__ == '__main__':
    run()



#!/usr/bin/env python
import os
from setuptools import setup, find_packages


long_description = open(
    os.path.join(
        os.path.dirname(__file__),
        'README.rst'
    )
).read()


setup(
    name='pymux',
    author='Jonathan Slenders',
    version='0.13',
    license='LICENSE',
    url='https://github.com/jonathanslenders/',
    description='Pure Python terminal multiplexer.',
    long_description=long_description,
    packages=find_packages('.'),
    install_requires = [
        'prompt_toolkit>=1.0.8,<1.1.0',
        'pyte>=0.5.1',
        'six>=1.9.0',
        'docopt>=0.6.2',
    ],
    entry_points={
        'console_scripts': [
            'pymux = pymux.entry_points.run_pymux:run',
        ]
    },
)
